# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import importlib
import inspect
from typing import Optional, List, Any
import yaml

from .context import Context


def _last_index(lst, element):
    """Get the last index of the element in a list.
    """
    return len(lst) - 1 - lst[::-1].index(element)


def instantiate_object(name: str, package: Optional[str] = None):
    """Instantiate an object defined in a Python source.

    Example:

        >>> instantiate_object('a.ClassA') # a ClassA type object from module a
        # a ClassB type object from module a.b
        >>> instantiate_object('..b.ClassB', package='a.c')

    Args:
        name (str): The name of the object. Support prefixing with additional package
            paths like `package_name.ObjectName`.
        package (str): The name of the package where the object is defined.

    Returns:
        Any: The instantiated object.

    """
    if '.' in name:
        # import module object first
        dot_index = _last_index(name, '.')
        module_name = name[:dot_index]
        class_name = name[dot_index+1:]
        module_object = importlib.import_module(module_name, package)
    else:
        if package is None:
            raise RuntimeError(
                f'A package name must be given to instantiate "{name}". '
                f'Otherwise, you should use a full class name that '
                f'contains its package path.')
        # `name` is a class name
        module_object = importlib.import_module(package)
        class_name = name
    if hasattr(module_object, class_name):
        return getattr(module_object, class_name)
    if hasattr(module_object, '__init__'):
        init = getattr(module_object, '__init__')
        if hasattr(init, class_name):
            return getattr(init, class_name)
    raise RuntimeError(
        f'Failed to find the class {class_name} in either '
        f'{module_object} or its __init__ file.')


def get_package_attributes(package: Optional[str]):
    """Get all attributes of a package.
    """
    if package is None:
        return dict()
    module_object = importlib.import_module(package)
    return module_object.__dict__


def _get_cfg_id(cfg_file: str, num_levels_kept: int) -> str:
    suffix = '.yaml'
    assert cfg_file.endswith(suffix)
    path_components = os.path.abspath(
        cfg_file[:-len(suffix)]).replace('\\', '/').split('/')
    path_components = path_components[-num_levels_kept:]
    cfg_id = '.'.join(path_components)
    return cfg_id


class Builder:
    """The builder that builds python objects from YAML files.

    The Builder extends an ordinary YAML parser with following extra functionality:

    1. A YAML dict with an entry `class: SomeClassName` will be parsed as a python object of
        class `SomeClassName`, all other entries will serve as the input arguments to 
        the `__init__` constructor of the class.

    2. A YAML dict with an entry `function: some_function_name` will be parsed as the result
        of the function `some_function_name`, with all other entries serving as the input arguments
        to that function.

    3. A string starting with `$` will be executed as a python script and returns the resulting python object. 
        Some variables or functions are pre-defined when executing these scripts:
          - `PARSE(relative_path_to_another_yaml_file, **kwargs)`: 
          a function that parses another YAML file with the same Builder and returns the parsed object. 
          For example, `$PARSE('../a.yaml', some_arg=some_value)`.
          - `ARGS`: a dict object that contains all the arguments that are fed into `PARSE`.
          - `FILE`: a string storing the path of this YAML file.
          - `RELATIVE(path)`: a function that returns a relative path to `FILE`.
          - `BLOB(path)`: a function that returns a path under the `blob_root` defined by context.
          - `CTX`: the `Context` object.

    4. Remember to include an entry of `package: package_name`, and import the releated classes and functions 
        in `package_name.__init__`, so that the builder can access their definitions for cases 1.-3.

    """
    def __init__(self, context: Context, verbose: bool = True, ignore_custom_parse: bool = False):
        self._ctx = context
        self.verbose = verbose
        self.ignore_custom_parse = ignore_custom_parse
        self._cfg_file_paths = [None]
        self._packages = [None]
        self._args_stack = [None]

    def _print(self, msg):
        if self.verbose:
            print(f'blueprint: {msg}')

    def parse_cfg_file(self, cfg_file_path, **kwargs):
        args = kwargs

        # xxx/xxx/xxx.yaml:a/b/c
        if ':' in cfg_file_path:
            cfg_file_path, inner_tree_path = cfg_file_path.split(':')
            inner_tree_path = inner_tree_path.split('/')
        else:
            inner_tree_path = []

        if self._cfg_file_paths[-1] is not None:
            cfg_file_path = os.path.join(
                os.path.dirname(self._cfg_file_paths[-1]), cfg_file_path)
        self._cfg_file_paths.append(cfg_file_path)
        self._args_stack.append(args)

        tree_data = yaml.safe_load(open(cfg_file_path, 'r'))
        self._print(f'Parsing {cfg_file_path}')

        if isinstance(tree_data, dict):
            package = tree_data.get('package', None)
        else:
            package = None
        for part in inner_tree_path:
            package = tree_data.get('package', package)
            tree_data = tree_data[part]
        if package is not None:
            self._packages.append(package)

        result = self.parse_cfg_object(tree_data)

        if package is not None:
            self._packages.pop()
        self._cfg_file_paths.pop()
        self._args_stack.pop()
        return result

    def blob_path(self, p):
        return os.path.join(self._ctx.blob_root, p)

    def states_path(self, cfg_file: str, num_levels_kept: int = 2):
        p = os.path.join(self._ctx.blob_root, 'states',
                         self._ctx.exp_name, _get_cfg_id(cfg_file, num_levels_kept))
        assert p.startswith(self._ctx.blob_root)
        return p

    def outputs_path(self, cfg_file: str, num_levels_kept: int = 2):
        p = os.path.join(self._ctx.blob_root, 'outputs',
                         self._ctx.exp_name, _get_cfg_id(cfg_file, num_levels_kept))
        assert p.startswith(self._ctx.blob_root)
        return p

    def relative_path(self, p: str):
        if self._cfg_file_paths[-1] is None:
            raise RuntimeError(
                f'No configuration file found, failed to compute the relative path of {p}.')
        return os.path.relpath(p, os.path.dirname(self._cfg_file_paths[-1]))

    def parse_cfg_object(self, cfg):
        ret = cfg

        if isinstance(cfg, str):
            if cfg.startswith('$$'):
                cfg = cfg.strip()
                arg_name = cfg[len('$$'):]
                ret = self._args_stack[-1][arg_name]

            elif cfg.startswith('$'):
                import math
                import numpy as np
                import torch

                try:
                    ret = eval(cfg[1:], dict(), {
                        'PARSE': self.parse_cfg_file,
                        'ARGS': self._args_stack[-1],
                        'FILE': self._cfg_file_paths[-1],
                        'RELATIVE': self.relative_path,
                        'BLOB': self.blob_path,
                        'STATES': self.states_path,
                        'OUTPUTS': self.outputs_path,
                        'CTX': self._ctx,
                        'os': os,
                        'math': math,
                        'np': np,
                        'torch': torch,
                        **get_package_attributes(self._packages[-1])
                    })
                except Exception as e:
                    self._print(f'Error when evaluating {cfg[1:]}')
                    self._print(f'Current package stack: {self._packages}')
                    raise

        elif isinstance(cfg, dict):
            if 'package' in cfg:
                package = cfg['package']
                del cfg['package']
                self._packages.append(package)
                ret = self.parse_cfg_object(cfg)
                self._packages.pop()
            else:
                if 'class' in cfg or 'function' in cfg:
                    instant_t = None
                    is_class = None
                    if 'class' in cfg:
                        instant_t = instantiate_object(
                            cfg['class'], self._packages[-1])
                        is_class = True
                    if 'function' in cfg:
                        if instant_t is not None:
                            raise RuntimeError(
                                'definition conflict: "class" and "function" '
                                'cannot appear under the same level')
                        instant_t = instantiate_object(
                            cfg['function'], self._packages[-1])
                        is_class = False

                    instant_t.__builder__ = self

                    input_args_dict = {key: val for key, val in cfg.items() if
                                       key not in {'class', 'function'}}

                    # update input_args_dict if the class/function has a '__parse_inputs__' member function
                    if hasattr(instant_t, '__parse_inputs__') and not self.ignore_custom_parse:
                        parse_inputs_fn = getattr(
                            instant_t, '__parse_inputs__')
                        input_args_to_parse = []
                        parse_inputs_fn_argnames = inspect.getfullargspec(
                            parse_inputs_fn).args
                        for arg_name in parse_inputs_fn_argnames:
                            input_args_to_parse.append(
                                input_args_dict[arg_name])
                        parsed_input_args = parse_inputs_fn(
                            *input_args_to_parse)
                        if not isinstance(parsed_input_args, tuple):
                            parsed_input_args = (parsed_input_args,)
                        if len(parsed_input_args) != len(parse_inputs_fn_argnames):
                            raise RuntimeError(
                                f'The __parse_inputs__ function should have the same number of inputs '
                                f'({len(parse_inputs_fn_argnames)}) and outputs ({len(parsed_input_args)})')
                    else:
                        parse_inputs_fn_argnames = []
                        parsed_input_args = []

                    for arg_name, parsed_arg in zip(parse_inputs_fn_argnames, parsed_input_args):
                        input_args_dict[arg_name] = parsed_arg
                    for arg_name, val in input_args_dict.items():
                        if arg_name not in parse_inputs_fn_argnames:
                            input_args_dict[arg_name] = self.parse_cfg_object(
                                val)

                    # check whether the required args of instant_t contains '_args'
                    if is_class:
                        # detect the args of __init__
                        instant_t_init_args = inspect.getfullargspec(
                            instant_t.__init__).args[1:]  # slice with [1:] to skip the `self` arg
                    else:
                        instant_t_init_args = inspect.getfullargspec(
                            instant_t).args

                    if '_ctx' in input_args_dict:
                        raise RuntimeError(
                            f'The argument "_ctx" is reserved and should not be assigned in `cfg`')

                    # if so, set the lastest args as '_args'
                    if '_ctx' in instant_t_init_args:
                        input_args_dict['_ctx'] = self._ctx

                    ret = instant_t(**input_args_dict)
                else:
                    ret = {name: self.parse_cfg_object(
                        val) for name, val in cfg.items()}

        elif isinstance(cfg, list):
            ret = [self.parse_cfg_object(val) for val in cfg]

        elif isinstance(cfg, tuple):
            ret = tuple([self.parse_cfg_object(val) for val in cfg])

        return ret


def build(context: Context, cfg_file_path: str, **kwargs):
    builder = Builder(context)
    return builder.parse_cfg_file(cfg_file_path, **kwargs)



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import functools
import operator
import numpy as np
import random
import cv2
import math
import functools

from typing import Tuple, Callable, Dict, List, Any, Union

from torch.nn.functional import interpolate

from .basic import AugBase
from .image import _to_cv_inter


def degrees(deg):
    return deg / 180.0 * math.pi


class GetShape(AugBase):
    """ Tensor -> [h, w, c, ...]
    """

    def process(self, image):
        return image.shape


class MatrixMul(AugBase):
    """ a, b, ... -> a @ b @ ... """

    def process(self, data):
        return functools.reduce(operator.matmul, data)


class MatrixInv(AugBase):
    """ a -> inv(a) """

    def process(self, m):
        return np.linalg.inv(m)


def _compose_rotate_and_scale(angle, scale, shift_xy, from_center, to_center):
    cosv = math.cos(angle)
    sinv = math.sin(angle)

    fx, fy = from_center
    tx, ty = to_center

    acos = scale * cosv
    asin = scale * sinv

    a0 = acos
    a1 = -asin
    a2 = tx - acos * fx + asin * fy + shift_xy[0]

    b0 = asin
    b1 = acos
    b2 = ty - asin * fx - acos * fy + shift_xy[1]

    rot_scale_m = np.array([
        [a0, a1, a2],
        [b0, b1, b2],
        [0.0, 0.0, 1.0]
    ], np.float32)
    return rot_scale_m


def _compose_flip_x(w):
    return np.array([
        [-1, 0, w-1],
        [0, 1, 0],
        [0, 0, 1]
    ], np.float32)


def _compose_flip_y(h):
    return np.array([
        [1, 0, 0],
        [0, -1, h-1],
        [0, 0, 1]
    ], np.float32)


class GetTransformMatrix(AugBase):
    """ from_shape -> [to_shape], transform_matrix
    """

    def __init__(self, target_shape,
                 shift_xy_mu=(0.0, 0.0), rot_mu=0.0, scale_mu=1.0,
                 flip_x=False, flip_y=False, ret_shape=False):
        self.target_shape = target_shape
        self.shift_xy_mu = shift_xy_mu
        self.rot_mu = rot_mu
        self.scale_mu = scale_mu
        self.flip_x = flip_x
        self.flip_y = flip_y
        self.ret_shape = ret_shape

    def process(self, from_shape):
        image_h, image_w, *_ = from_shape
        assert isinstance(image_h, int)
        assert isinstance(image_w, int)

        to_h, to_w, *_ = self.target_shape

        matrix = _compose_rotate_and_scale(
            self.rot_mu, self.scale_mu, self.shift_xy_mu,
            from_center=[(image_w-1)/2.0, (image_h-1)/2.0],
            to_center=[(to_w-1)/2.0, (to_h-1)/2.0])
        if self.flip_x:
            matrix = _compose_flip_x(to_w) @ matrix
        if self.flip_y:
            matrix = _compose_flip_y(to_h) @ matrix
        if self.ret_shape:
            return (to_h, to_w), matrix
        return matrix


class UpdateTransformMatrix(AugBase):
    """ from_shape, transform_matrix -> [to_shape], transform_matrix
    """

    def __init__(self, target_shape,
                 shift_xy_mu=(0.0, 0.0), rot_mu=0.0, scale_mu=1.0,
                 flip_x=False, flip_y=False, ret_shape=False):
        self.target_shape = target_shape
        self.shift_xy_mu = shift_xy_mu
        self.rot_mu = rot_mu
        self.scale_mu = scale_mu
        self.flip_x = flip_x
        self.flip_y = flip_y
        self.ret_shape = ret_shape

    def process(self, data):
        from_shape, matrix = data

        image_h, image_w, *_ = from_shape
        to_h, to_w, *_ = self.target_shape

        trans_matrix = _compose_rotate_and_scale(
            self.rot_mu, self.scale_mu, self.shift_xy_mu,
            from_center=[(image_w-1)/2.0, (image_h-1)/2.0],
            to_center=[(to_w-1)/2.0, (to_h-1)/2.0])

        if matrix is not None:
            matrix = trans_matrix @ matrix
        else:
            matrix = trans_matrix

        if self.flip_x:
            matrix = _compose_flip_x(to_w) @ matrix
        if self.flip_y:
            matrix = _compose_flip_y(to_h) @ matrix
        if self.ret_shape:
            return (to_h, to_w), matrix
        return matrix


class UpdateRandomTransformMatrix(AugBase):
    """ from_shape, transform_matrix -> [to_shape], transform_matrix
    """

    def __init__(self, target_shape,
                 shift_sigma=0.1, rot_sigma=degrees(18), scale_sigma=0.1,
                 shift_mu=0.0, rot_mu=0.0, scale_mu=1.0,
                 shift_normal=True, rot_normal=True, scale_normal=True,
                 flip_x_rand=False, flip_y_rand=False, ret_shape=False):

        self.target_shape = target_shape
        self.shift_config = (shift_mu, shift_sigma, shift_normal)
        self.rot_config = (rot_mu, rot_sigma, rot_normal)
        self.scale_config = (scale_mu, scale_sigma, scale_normal)
        self.flip_x_rand = flip_x_rand
        self.flip_y_rand = flip_y_rand
        self.ret_shape = ret_shape

    def _random(self, mu_sigma_normal, size=None):
        mu, sigma, is_normal = mu_sigma_normal
        if is_normal:
            return np.random.normal(mu, sigma, size=size)
        else:
            return np.random.uniform(low=mu-sigma, high=mu+sigma, size=size)

    def process(self, data):
        assert isinstance(data, (tuple, list))
        from_shape, matrix = data
        assert isinstance(matrix, np.ndarray)

        from_h, from_w, *_ = from_shape
        to_h, to_w, *_ = self.target_shape

        if self.shift_config[:2] != (0.0, 0.0) or \
           self.rot_config[:2] != (0.0, 0.0) or \
           self.scale_config[:2] != (1.0, 0.0):
            shift_xy = self._random(self.shift_config, size=[2]) * \
                min(to_h, to_w)
            rot_angle = self._random(self.rot_config)
            scale = self._random(self.scale_config)
            trans_matrix = _compose_rotate_and_scale(
                rot_angle, scale, shift_xy,
                from_center=[(from_w-1)/2.0, (from_h-1)/2.0],
                to_center=[(to_w-1)/2.0, (to_h-1)/2.0])
            matrix = trans_matrix @ matrix

        if self.flip_x_rand and bool(random.getrandbits(1)):
            matrix = _compose_flip_x(to_w) @ matrix

        if self.flip_y_rand and bool(random.getrandbits(1)):
            matrix = _compose_flip_y(to_h) @ matrix

        if self.ret_shape:
            return (to_h, to_w), matrix
        return matrix


class UpdateResizeMatrix(AugBase):
    """ from_shape, transform_matrix -> [to_shape], transform_matrix
    """

    def __init__(self, target_shape, align_corners, ret_shape=False):
        self.target_shape = target_shape
        self.align_corners = align_corners
        self.ret_shape = ret_shape

    def process(self, data):
        (from_h, from_w, *_), matrix = data

        to_h, to_w, *_ = self.target_shape

        if self.align_corners:
            # to_x = from_x/(from_w-1)*(to_w-1)
            trans_matrix = np.array([
                [(to_w-1)/(from_w-1), 0, 0],
                [0, (to_h-1)/(from_h-1), 0],
                [0, 0, 1]], dtype=matrix.dtype)
        else:
            # to_x = (from_x+0.5)/from_w*to_w-0.5
            trans_matrix = np.array([
                [to_w/from_w, 0, 0.5*to_w/from_w-0.5],
                [0, to_h/from_h, 0.5*to_h/from_h-0.5],
                [0, 0, 1]], dtype=matrix.dtype)

        if matrix is None:
            matrix = trans_matrix
        else:
            matrix = trans_matrix @ matrix

        if self.ret_shape:
            return (to_h, to_w), matrix
        return matrix


class CenterSquaredBoxWithMaxSize(AugBase):
    """ from_shape -> box (y1x1y2x2)
    """

    def process(self, from_shape):
        h, w = from_shape[:2]
        m = min(h, w)
        y1 = (h-m) * 0.5
        x1 = (w-m) * 0.5
        return np.array([y1, x1, y1+m, x1+m], dtype=np.float32)


class RandomSquaredBoxWithMaxSize(AugBase):
    """ from_shape -> box (y1x1y2x2)
    """

    def process(self, from_shape):
        h, w = from_shape[:2]
        m = min(h, w)
        y1 = np.random.rand() * (h-m)
        x1 = np.random.rand() * (w-m)
        return np.array([y1, x1, y1+m, x1+m], dtype=np.float32)


class RandomSquaredBoxWithRandomSize(AugBase):
    """ from_shape -> box (y1x1y2x2)
    """

    def __init__(self, min_size, max_size) -> None:
        self.min_size = min_size
        self.max_size = max_size

    def process(self, from_shape):
        h, w = from_shape[:2]
        m = min(h, w)
        min_size = min(self.min_size, m)
        max_size = min(self.max_size, m)
        size = np.random.rand() * (max_size-min_size) + min_size
        y1 = np.random.rand() * (h-size)
        x1 = np.random.rand() * (w-size)
        return np.array([y1, x1, y1+size, x1+size], dtype=np.float32)


class UpdateCropAndResizeMatrix(AugBase):
    """ box (y1x1y2x2), matrix (3x3) -> [to_shape], matrix (3x3)

    Args:
        align_corners (bool): Set this to `True` only if the box you give has coordinates
            ranging from `0` to `h-1` or `w-1`.

        offset_box_coords (bool): Set this to `True` if the box you give has coordinates
            ranging from `0` to `h` or `w`. 

            Set this to `False` if the box coordinates range from `-0.5` to `h-0.5` or `w-0.5`.

            If the box coordinates range from `0` to `h-1` or `w-1`, set `align_corners=True`.
    """

    def __init__(self, target_shape, align_corners, ret_shape=False,
                 offset_box_coords=False):
        self.target_shape = target_shape
        self.align_corners = align_corners
        self.ret_shape = ret_shape
        self.offset_box_coords = offset_box_coords

    def process(self, data):
        (y1, x1, y2, x2), matrix = data
        h, w, *_ = self.target_shape
        dtype = matrix.dtype if matrix is not None else np.float32
        if self.align_corners:
            # x -> (x - x1) / (x2 - x1) * (w - 1)
            # y -> (y - y1) / (y2 - y1) * (h - 1)
            ax = 1.0 / (x2 - x1) * (w - 1)
            ay = 1.0 / (y2 - y1) * (h - 1)
            trans_matrix = np.array([
                [ax, 0, -x1 * ax],
                [0, ay, -y1 * ay],
                [0, 0, 1]
            ], dtype=dtype)
        else:
            if self.offset_box_coords:
                # x1, x2 \in [0, w], y1, y2 \in [0, h]
                # first we should offset x1, x2, y1, y2 to be ranging in
                # [-0.5, w-0.5] and [-0.5, h-0.5]
                # so to convert these pixel coordinates into boundary coordinates.
                x1, x2, y1, y2 = x1-0.5, x2-0.5, y1-0.5, y2-0.5

            # x -> (x - x1) / (x2 - x1) * w - 0.5
            # y -> (y - y1) / (y2 - y1) * h - 0.5
            ax = 1.0 / (x2 - x1) * w
            ay = 1.0 / (y2 - y1) * h
            trans_matrix = np.array([
                [ax, 0, -x1 * ax - 0.5],
                [0, ay, -y1 * ay - 0.5],
                [0, 0, 1]
            ], dtype=dtype)

        if matrix is None:
            matrix = trans_matrix
        else:
            matrix = trans_matrix @ matrix

        if self.ret_shape:
            return (h, w), matrix
        return matrix


def _make_n2p_mat(shape, align_corners):
    if shape is None:
        return None
    h, w = shape[:2]
    if align_corners:
        h, w = h-1, w-1
    return np.array([
        [w/2, 0, w/2],
        [0, h/2, h/2],
        [0, 0, 1]
    ], dtype=np.float32)


class ConvertTransformMatrixToPyTorchTheta(AugBase):
    """ ConvertTransformMatrixToPyTorchTheta

        from_shape, transform_matrix (3x3) -> theta (2x3)

        Convert a 3x3 transform matrix into a 2x3 theta matrix for 
        invoking torch.nn.functional.affine_grid.
    """

    def __init__(self, target_shape, inverse=True, keep3x3=False, align_corners=False):
        self.p2n_to = np.linalg.inv(_make_n2p_mat(target_shape, align_corners))
        self.inverse = inverse
        self.keep3x3 = keep3x3
        self.align_corners = align_corners

    def process(self, data):
        from_shape, matrix = data
        n2p_from = _make_n2p_mat(from_shape, self.align_corners)
        if self.inverse:
            matrix = np.linalg.inv(matrix)
        theta = self.p2n_to @ matrix @ n2p_from
        if self.keep3x3:
            return theta
        else:
            return theta[:2, :]


class ConvertPyTorchThetaToTransformMatrix(AugBase):
    """ ConvertTransformMatrixToPyTorchTheta

        from_shape, theta (2x3) -> transform_matrix (3x3)

        Convert a 2x3 theta matrix into a 3x3 transform matrix.
    """

    def __init__(self, target_shape, inverse=True, align_corners=False):
        self.n2p_to = _make_n2p_mat(target_shape, align_corners)
        self.inverse = inverse
        self.align_corners = align_corners

    def process(self, data):
        from_shape, theta = data
        p2n_from = np.linalg.inv(_make_n2p_mat(from_shape, self.align_corners))

        if theta.size(0) == 2:
            theta = np.concatenate(
                [theta, np.zeros([1, 3], dtype=theta.dtype)], axis=0)

        matrix = self.n2p_to @ theta @ p2n_from
        if self.inverse:
            matrix = np.linalg.inv(matrix)
        return matrix


@functools.lru_cache(maxsize=128)
def _meshgrid(h, w):
    yy, xx = np.meshgrid(np.arange(0, h, dtype=np.float32),
                         np.arange(0, w, dtype=np.float32),
                         indexing='ij')
    return yy, xx


def _forge_transform_map(output_shape, fn: Callable[[np.ndarray], np.ndarray]
                         ) -> Tuple[np.ndarray, np.ndarray]:
    """ Forge transform maps with a given function `fn`.

    Args:
        output_shape (tuple): (h, w, ...).
        fn (Callable[[np.ndarray], np.ndarray]): The function that accepts 
            a Nx2 array and outputs the transformed Nx2 array. Both input 
            and output store (x, y) coordinates.

    Note: 
        both input and output arrays of `fn` should store (y, x) coordinates.

    Returns:
        Tuple[np.ndarray, np.ndarray]: Two maps `X` and `Y`, where for each pixel (y, x) or coordinate (x, y),
            `(X[y, x], Y[y, x]) = fn([x, y])`
    """
    h, w, *_ = output_shape
    yy, xx = _meshgrid(h, w)  # h x w
    in_xxyy = np.stack([xx, yy], axis=-1).reshape([-1, 2])  # (h x w) x 2
    out_xxyy: np.ndarray = fn(in_xxyy)  # (h x w) x 2
    return out_xxyy.reshape([h, w, 2])


def _safe_arctanh(x):
    x[x < -0.999] = -0.999
    x[x > +0.999] = +0.999
    x = np.arctanh(x)
    return x


def _tanh_warp_transform(coords, transform_matrix, warp_factor, warped_shape):
    """ Tanh-warp function.

    Args:
        coords (np.ndarray): N x 2 (x, y).
        transform_matrix: 3 x 3. A matrix that transforms un-normalized coordinates 
            from the original image to the aligned yet not-warped image.
        warp_factor (float): The warp factor. 
            0 means linear transform, 1 means full tanh warp.
        warped_shape (tuple): [height, width].

    Returns:
        np.ndarray: N x 2 (x, y).
    """
    h, w, *_ = warped_shape
    # h -= 1
    # w -= 1

    if warp_factor > 0:
        # normalize coordinates to [-1, +1]
        coords = coords / np.array([w, h], dtype=coords.dtype) * 2 - 1

        nl_part1 = coords > 1.0 - warp_factor
        nl_part2 = coords < -1.0 + warp_factor

        coords[nl_part1] = _safe_arctanh(
            (coords[nl_part1] - 1.0 + warp_factor) /
            warp_factor) * warp_factor + \
            1.0 - warp_factor
        coords[nl_part2] = _safe_arctanh(
            (coords[nl_part2] + 1.0 - warp_factor) /
            warp_factor) * warp_factor - \
            1.0 + warp_factor

        # denormalize
        coords = (coords + 1) / 2 * np.array([w, h], dtype=coords.dtype)

    coords_homo = np.concatenate(
        [coords, np.ones([coords.shape[0], 1], dtype=coords.dtype)], axis=1)  # N x 3

    inv_matrix = np.linalg.inv(transform_matrix)
    coords_homo = np.dot(coords_homo, inv_matrix.T)  # N x 3
    return (coords_homo[:, :2] / coords_homo[:, [2, 2]]).astype(coords.dtype)


def _inverted_tanh_warp_transform(coords, transform_matrix, warp_factor, warped_shape):
    """ Inverted Tanh-warp function.

    Args:
        coords (np.ndarray): N x 2 (x, y).
        transform_matrix: 3 x 3. A matrix that transforms un-normalized coordinates 
            from the original image to the aligned yet not-warped image.
        warp_factor (float): The warp factor. 
            0 means linear transform, 1 means full tanh warp.
        warped_shape (tuple): [height, width].

    Returns:
        np.ndarray: N x 2 (x, y).
    """
    h, w, *_ = warped_shape
    # h -= 1
    # w -= 1

    coords_homo = np.concatenate(
        [coords, np.ones([coords.shape[0], 1], dtype=coords.dtype)], axis=1)  # N x 3

    coords_homo = np.dot(coords_homo, transform_matrix.T)  # N x 3
    coords = (coords_homo[:, :2] / coords_homo[:, [2, 2]]).astype(coords.dtype)

    if warp_factor > 0:
        # normalize coordinates to [-1, +1]
        coords = coords / np.array([w, h], dtype=coords.dtype) * 2 - 1

        nl_part1 = coords > 1.0 - warp_factor
        nl_part2 = coords < -1.0 + warp_factor

        coords[nl_part1] = np.tanh(
            (coords[nl_part1] - 1.0 + warp_factor) /
            warp_factor) * warp_factor + \
            1.0 - warp_factor
        coords[nl_part2] = np.tanh(
            (coords[nl_part2] + 1.0 - warp_factor) /
            warp_factor) * warp_factor - \
            1.0 + warp_factor

        # denormalize
        coords = (coords + 1) / 2 * np.array([w, h], dtype=coords.dtype)

    return coords


class GetTransformMap(AugBase):
    """ transform_matrix -> map

    Args:
        warped_shape: The target image shape to transform to.
        warp_factor: The warping factor. `warp_factor=1.0` represents a vannila Tanh-warping, 
            `warp_factor=0.0` represents a cropping.

    This involves an implementation of the Tanh-warping, which is proposed in:

        [1] Lin, Jinpeng, Hao Yang, Dong Chen, Ming Zeng, Fang Wen, and Lu Yuan. 
            "Face parsing with roi tanh-warping." In Proceedings of the IEEE/CVF 
            Conference on Computer Vision and Pattern Recognition, pp. 5654-5663. 2019.

        [2] Zheng, Yinglin, Hao Yang, Ting Zhang, Jianmin Bao, Dongdong Chen, Yangyu Huang,
            Lu Yuan, Dong Chen, Ming Zeng and Fang Wen. "General Facial Representation 
            Learning in a Visual-Linguistic Manner" In arxiv.

    Please cite the paper on your usage.

    """

    def __init__(self, warped_shape, warp_factor=0.0):
        self.warped_shape = warped_shape
        self.warp_factor = warp_factor

    def process(self, matrix):
        return _forge_transform_map(
            self.warped_shape,
            functools.partial(_tanh_warp_transform,
                              transform_matrix=matrix,
                              warp_factor=self.warp_factor,
                              warped_shape=self.warped_shape))


class GetInvertedTransformMap(AugBase):
    """ transform_matrix, original_image_shape -> inverted_map

    This is an inverted transform of GetTransformMap.

    Args:
        warped_shape: The target image shape to transform to.
        warp_factor: The warping factor. `warp_factor=1.0` represents a vannila Tanh-warping, 
            `warp_factor=0.0` represents a cropping.

    This involves an implementation of the Tanh-warping, which is proposed in:

        [1] Lin, Jinpeng, Hao Yang, Dong Chen, Ming Zeng, Fang Wen, and Lu Yuan. 
            "Face parsing with roi tanh-warping." In Proceedings of the IEEE/CVF 
            Conference on Computer Vision and Pattern Recognition, pp. 5654-5663. 2019.
            
        [2] Zheng, Yinglin, Hao Yang, Ting Zhang, Jianmin Bao, Dongdong Chen, Yangyu Huang,
            Lu Yuan, Dong Chen, Ming Zeng and Fang Wen. "General Facial Representation 
            Learning in a Visual-Linguistic Manner" In arxiv.

    Please cite the paper on your usage.

    """

    def __init__(self, warped_shape, warp_factor=0.0):
        self.warped_shape = warped_shape
        self.warp_factor = warp_factor

    def process(self, data):
        matrix, orig_shape = data
        return _forge_transform_map(
            orig_shape,
            functools.partial(_inverted_tanh_warp_transform,
                              transform_matrix=matrix,
                              warp_factor=self.warp_factor,
                              warped_shape=self.warped_shape))


class TransformByMap(AugBase):
    """ image, map -> transformed_image
    """

    def __init__(self, interpolation: str, outlier_value=0, outlier_range=None):
        self.interpolation = _to_cv_inter(interpolation)
        self.outlier_value = outlier_value
        self.outlier_range = outlier_range

    def process(self, data):
        image, map_xxyy = data
        xx = map_xxyy[:, :, 0]
        yy = map_xxyy[:, :, 1]
        if self.outlier_range is None:
            ov = self.outlier_value
        else:
            ov = np.random.uniform(
                low=self.outlier_range[0], high=self.outlier_range[1],
                size=[image.shape[2]])
        return cv2.remap(image, xx, yy, self.interpolation,
                         borderMode=cv2.BORDER_CONSTANT,
                         borderValue=ov)


class TransformPoints2D(AugBase):
    """ points, matrix3x3 -> points
    """

    def __init__(self, warped_shape, warp_factor=0.0):
        self.warped_shape = warped_shape
        self.warp_factor = warp_factor

    def process(self, data):
        points, matrix = data
        return _inverted_tanh_warp_transform(
            points, matrix, self.warp_factor, self.warped_shape)


class TransformPoints2DInverted(AugBase):
    """ points, matrix3x3 -> points
    """

    def __init__(self, warped_shape, warp_factor=0.0):
        self.warped_shape = warped_shape
        self.warp_factor = warp_factor

    def process(self, data):
        points, matrix = data
        return _tanh_warp_transform(
            points, matrix, self.warp_factor, self.warped_shape)


class TransformAll(AugBase):
    """ (image/point, ...), matrix3x3 -> (transformed_image/transformed_point, ...)
    """

    def __init__(self, warped_shape, warp_factor: float,
                 warp_configs: Union[Dict[str, Any], List[Dict[str, Any]]]) -> None:
        self.warped_shape = warped_shape
        self.warp_factor = warp_factor

        if isinstance(warp_configs, dict):
            warp_configs = [warp_configs]
            self.is_single_input = True
        else:
            self.is_single_input = False
        self.warp_configs = warp_configs
        self.has_image = any(
            [cfg['data_type'] == 'image' for cfg in self.warp_configs])

    def process(self, data):
        input_data, matrix = data
        if self.is_single_input:
            input_data = [input_data]

        outputs = []
        if self.has_image:
            map_xxyy = _forge_transform_map(
                self.warped_shape,
                functools.partial(_tanh_warp_transform,
                                  transform_matrix=matrix,
                                  warp_factor=self.warp_factor,
                                  warped_shape=self.warped_shape))
            xx = map_xxyy[:, :, 0]
            yy = map_xxyy[:, :, 1]

        for d, cfg in zip(input_data, self.warp_configs):
            assert isinstance(d, np.ndarray)
            data_type = cfg['data_type']
            if data_type == 'image':
                assert d.ndim == 3
                if 'outlier_range' not in cfg:
                    ov = cfg.get('outlier_value', 0)
                else:
                    ov = np.random.uniform(
                        low=cfg['outlier_range'][0], high=cfg['outlier_range'][1],
                        size=[d.shape[2]])
                outputs.append(cv2.remap(d, xx, yy, _to_cv_inter(cfg['interpolation']),
                                         borderMode=cv2.BORDER_CONSTANT,
                                         borderValue=ov))
            elif data_type == 'point':
                assert d.ndim == 2
                assert d.shape[1] == 2
                outputs.append(_inverted_tanh_warp_transform(
                    d, matrix, self.warp_factor, self.warped_shape))
            else:
                raise RuntimeError(
                    f'Unsupported data_type in TransformWithWarp: {data_type}')

        if self.is_single_input:
            return outputs[0]
        return outputs


class TransformPoints2DPerspective(AugBase):
    """ points (nx2), matrix3x3 -> points (nx2)
    """

    def process(self, data):
        points, matrix = data
        dtype = points.dtype

        # nx3
        points = np.concatenate([points, np.ones_like(points[:, [0]])], axis=1)
        points = points @ np.transpose(matrix)  # nx3
        points = points[:, :2] / points[:, [2, 2]]
        return points.astype(dtype)


class TransformImagePerspective(AugBase):
    """ image, matrix3x3 -> transformed_image
    """

    def __init__(self, target_shape, interpolation: str, outlier_value=0, outlier_range=None):
        self.target_shape = target_shape
        self.interpolation = _to_cv_inter(interpolation)
        self.outlier_value = outlier_value
        self.outlier_range = outlier_range

    def process(self, data):
        image, mat = data
        if self.outlier_range is None:
            ov = self.outlier_value
        else:
            ov = np.random.uniform(
                low=self.outlier_range[0], high=self.outlier_range[1],
                size=[image.shape[2]])
        return cv2.warpPerspective(
            image, mat, dsize=(self.target_shape[1], self.target_shape[0]),
            flags=self.interpolation, borderValue=ov)


class FlipLROrderedPoints(AugBase):
    """ points (nx2) -> points (nx2)
    """

    def __init__(self, image_shape, flip_mapping: List[Tuple[int, int]]) -> None:
        self.image_shape = image_shape
        self.flip_mapping = flip_mapping

    def process(self, points):
        _, w, *_ = self.image_shape
        points_flip = np.array(points)
        for i, j in self.flip_mapping:
            points_flip[i] = points[j]
            points_flip[j] = points[i]
        points_flip[:, 0] = w - 1 - points_flip[:, 0]
        return points_flip



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Callable, Any

import numpy as np
import functools
from skimage import transform

from .basic import AugBase

try:
    from face_sdk import FaceDetection
except ImportError:
    FaceDetection = None


def _compute_iou(boxA, boxB):
    """compute_iou
    y1, x1, y2, x2
    """
    # determine the (x, y)-coordinates of the intersection rectangle
    xA = max(boxA[1], boxB[1])
    yA = max(boxA[0], boxB[0])
    xB = min(boxA[3], boxB[3])
    yB = min(boxA[2], boxB[2])

    # compute the area of intersection rectangle
    interArea = max(0, xB - xA) * max(0, yB - yA)

    # compute the area of both the prediction and ground-truth
    # rectangles
    boxAArea = (boxA[3] - boxA[1]) * (boxA[2] - boxA[0])
    boxBArea = (boxB[3] - boxB[1]) * (boxB[2] - boxB[0])

    # compute the intersection over union by taking the intersection
    # area and dividing it by the sum of prediction + ground-truth
    # areas - the interesection area
    iou = interArea / float(boxAArea + boxBArea - interArea)

    # return the intersection over union value
    return iou


def _get_pixel(single_mask, y, x, outlier=0):
    h = single_mask.shape[0]
    w = single_mask.shape[1]
    if y < 0 or y >= h or x < 0 or x >= w:
        return outlier
    return single_mask[y, x]


def _get_target_fa_points(label_names, label, fa_points):
    if fa_points.size == 0:
        return fa_points

    max_num_matches = 0
    num_faces = fa_points.shape[0]
    best_face_id = -1
    for ii in range(num_faces):
        le_x, le_y, re_x, re_y, nose_x, nose_y, *_ = [
            int(fa_points[ii, k, t]) for k in range(5) for t in range(2)]
        num_matches = 0
        for px, py, names in [
            (le_x, le_y, ['le', 'face', 'face_neck', 'fg']),
            (re_x, re_y, ['re', 'face', 'fg']),
                (nose_x, nose_y, ['nose', 'face', 'fg'])]:
            for name in names:
                if name in label_names and _get_pixel(
                        label, py, px, -1) == label_names.index(name):
                    num_matches += 1

        if num_matches > max_num_matches:
            best_face_id = ii
            max_num_matches = num_matches

    fa_points = fa_points[best_face_id]
    return fa_points


def _filter_fa_points(face_rects, fa_points,
                      area_thres=12, overlap_thres=0.5):
    if face_rects.shape[0] == 0:
        return face_rects, fa_points

    face_weights = (face_rects[:, 3] - face_rects[:, 1]) * \
        (face_rects[:, 2] - face_rects[:, 0])
    valid_face_inds = face_weights > area_thres
    face_weights = face_weights[valid_face_inds]
    face_rects = face_rects[valid_face_inds]
    fa_points = fa_points[valid_face_inds]

    # remove fa_points with high overlaps
    inds = np.argsort(face_weights)[::-1]
    kept_inds = []
    for ind in inds:
        rect = face_rects[ind, :]
        overlapped = False
        for kind in kept_inds:
            krect = face_rects[kind, :]
            if _compute_iou(rect, krect) > overlap_thres:
                overlapped = True
                break
        if not overlapped:
            kept_inds.append(ind)
    face_weights = face_weights[kept_inds]
    face_rects = face_rects[kept_inds, :]
    fa_points = fa_points[kept_inds, :, :]
    return face_rects, fa_points


def _make_face_detector(detector) -> FaceDetection:
    if FaceDetection is None:
        raise RuntimeError('FaceSDK is not installed')
    if detector == 'jda':
        return FaceDetection(algorithm_version=1)
    if detector == 'dnn':
        return FaceDetection(algorithm_version=5)
    if detector == 'sln':
        return FaceDetection(algorithm_version=11)
    if isinstance(detector, int):
        return FaceDetection(algorithm_version=detector)
    raise RuntimeError('unsupported detector type')


_default_detector_ = 'dnn'


def _convert_face_rects(face_rects_xywh):
    xs, ys, ws, hs = np.split(face_rects_xywh, 4, axis=1)
    x2s = xs + ws
    y2s = ys + hs
    return np.concatenate([ys, xs, y2s, x2s], axis=1)


class DetectFaces(AugBase):
    """ image -> multi_fa_points, [face_rects]
    """

    def __init__(self,
                 area_thres=12, overlap_thres=0.5, max_faces=1000,
                 detector=_default_detector_, ret_rects=False):
        self.det = _make_face_detector(detector)
        self.area_thres = area_thres
        self.overlap_thres = overlap_thres
        self.max_faces = max_faces
        self.ret_rects = ret_rects

    def process(self, image):
        face_rects_xywh, multi_fa_points = self.det.detect_and_align(image)
        multi_fa_points = np.reshape(multi_fa_points[:, :10], [-1, 5, 2])
        face_rects, multi_fa_points = _filter_fa_points(
            _convert_face_rects(face_rects_xywh)[:self.max_faces],
            multi_fa_points[:self.max_faces],
            area_thres=self.area_thres,
            overlap_thres=self.overlap_thres)

        if self.ret_rects:
            return multi_fa_points, face_rects
        else:
            return multi_fa_points


class DetectFace(AugBase):
    """ detect single face
        image -> fa_points, [face_rect]
    """

    def __init__(self, detector=_default_detector_, ret_rect=False):
        self.det = _make_face_detector(detector)
        self.ret_rect = ret_rect

    def process(self, image):
        face_rects_xywh, multi_fa_points = self.det.detect_and_align(image)
        multi_fa_points = np.reshape(multi_fa_points[:, :10], [-1, 5, 2])
        face_rects = _convert_face_rects(face_rects_xywh)
        if self.ret_rect:
            return multi_fa_points[0], face_rects[0]
        else:
            return multi_fa_points[0]


class DetectFaceWithGroundtruthScorer(DetectFaces):
    """ image, gt_data -> fa_points, [face_rect]
    """

    def __init__(self, scorer: Callable[[np.ndarray, Any], np.ndarray],
                 area_thres=12, overlap_thres=0.5, max_faces=1000,
                 detector=_default_detector_, ret_rects=False):
        super().__init__(
            area_thres, overlap_thres, max_faces, detector, ret_rects)
        self.scorer = scorer

    def process(self, data):
        image, gt_data = data
        if self.ret_rects:
            multi_fa_points, face_rects = super().process(image)
        else:
            multi_fa_points = super().process(image)

        face_gt_scores = self.scorer(multi_fa_points, gt_data)
        max_ind = np.argmax(face_gt_scores)

        if self.ret_rects:
            return multi_fa_points[max_ind], face_rects[max_ind]
        else:
            return multi_fa_points[max_ind]


class _ComputeFaceGtHits:
    def __init__(self, label_names):
        self.label_names = label_names

    def __call__(self, multi_fa_points, label):
        num_faces = multi_fa_points.shape[0]
        face_gt_hits = np.zeros((num_faces,), dtype=np.uint8)
        for ii in range(num_faces):
            le_x, le_y, re_x, re_y, nose_x, nose_y, *_ = [
                int(multi_fa_points[ii, k, t]) for k in range(5) for t in range(2)]
            for px, py, names in [
                (le_x, le_y, ['le', 'face', 'face_neck', 'fg']),
                (re_x, re_y, ['re', 'face', 'fg']),
                    (nose_x, nose_y, ['nose', 'face', 'fg'])]:
                for name in names:
                    if name in self.label_names and _get_pixel(
                            label, py, px, -1) == self.label_names.index(name):
                        face_gt_hits[ii] += 1
        return face_gt_hits


class DetectFaceWithGroundtruthLabel(DetectFaceWithGroundtruthScorer):
    """ image, label -> fa_points, [face_rect]
    """

    def __init__(self, label_names,
                 area_thres=12, overlap_thres=0.5, max_faces=1000,
                 detector=_default_detector_, ret_rects=False):
        super().__init__(
            _ComputeFaceGtHits(label_names),
            area_thres, overlap_thres, max_faces, detector, ret_rects)


@functools.lru_cache()
def _standard_face_pts():
    pts = np.array([
        196.0, 226.0,
        316.0, 226.0,
        256.0, 286.0,
        220.0, 360.4,
        292.0, 360.4], np.float32) / 256.0 - 1.0
    return np.reshape(pts, (5, 2))


class GetFaceAlignMatrices(AugBase):
    """ multi_face_align_points -> [out_shape], tranform_matrices
        or
        single_face_align_points -> [out_shape], transform_matrix
    """

    def __init__(self, target_shape, target_face_scale=1.0,
                 offset_xy=None, target_pts=None, ret_shape=False):
        if target_pts is None:
            std_pts = _standard_face_pts()  # [-1 1]
            h, w, *_ = target_shape
            self.target_pts = (std_pts * target_face_scale + 1) * \
                np.array([w-1, h-1], np.float32) / 2.0
            if offset_xy is not None:
                self.target_pts[:, 0] += offset_xy[0]
                self.target_pts[:, 1] += offset_xy[1]
        else:
            self.target_pts = np.array(target_pts)
        self.ret_shape = ret_shape

    def _estimate_single_face_align_matrix(self, fa_points):
        tform = transform.SimilarityTransform()
        tform.estimate(fa_points, self.target_pts)
        return tform.params

    def process(self, multi_fa_points):
        assert multi_fa_points.shape[-2:] == (5, 2)
        if multi_fa_points.ndim == 3:
            matrix = np.stack([self._estimate_single_face_align_matrix(fa_points)
                               for fa_points in multi_fa_points], 0)
        else:
            matrix = self._estimate_single_face_align_matrix(multi_fa_points)
        if self.ret_shape:
            return self.target_shape, matrix
        return matrix


GetFaceAlignMatrix = GetFaceAlignMatrices


class GetFaceRandomCropBox(AugBase):
    """ imshape, fa_points -> box
    """

    def __init__(self, crop_rescale_mean, crop_rescale_var,
                 crop_rescale_min=0.1, crop_rescale_max=10):
        self.crop_rescale_stats = (
            crop_rescale_mean, crop_rescale_var, crop_rescale_min, crop_rescale_max)

    def process(self, data):
        crop_rescale_mean, crop_rescale_var,\
            crop_rescale_min, crop_rescale_max = self.crop_rescale_stats
        crop_rescale = np.random.normal(
            crop_rescale_mean, crop_rescale_var)
        if crop_rescale < crop_rescale_min:
            crop_rescale = crop_rescale_min
        if crop_rescale > crop_rescale_max:
            crop_rescale = crop_rescale_max
        crop_dx_dy = np.random.uniform(0, 1, size=2)

        imshape, fa_points = data
        assert fa_points.shape == (5, 2)
        face_size = np.linalg.norm(fa_points[0, :] - fa_points[1, :])

        rescaled_im = int(face_size * crop_rescale)
        rescaled_im = min([imshape[0], imshape[1], rescaled_im])

        rand_dx, rand_dy = crop_dx_dy
        rand_dx = int(rand_dx * (imshape[1] - rescaled_im))
        rand_dy = int(rand_dy * (imshape[0] - rescaled_im))
        return np.array([rand_dy, rand_dx, rand_dy + rescaled_im,
                         rand_dx + rescaled_im], dtype=np.int32)



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import time
import os
from typing import List
from collections import defaultdict

import torch
from torch import distributed as dist
import torchvision

from ..context import Context


def _tagged(s: str, tag) -> str:
    if tag is None:
        return s
    return f'{tag}.{s}'


class Logger:
    def setup(self, outputs_dir):
        pass

    def log_info(self, msg):
        pass

    def log_batched_images(self, outputs, global_step=None, epoch=None, tag=None):
        pass

    def log_numbers(self, numbers: dict, global_step=None, epoch=None, tag=None):
        pass


class Loggers(Logger):
    def __init__(self, loggers: List[Logger]) -> None:
        self.loggers = loggers

    def setup(self, outputs_dir):
        for logger in self.loggers:
            logger.setup(outputs_dir)

    def log_info(self, msg):
        for logger in self.loggers:
            logger.log_info(msg)

    def log_batched_images(self, outputs, global_step=None, epoch=None, tag=None):
        for logger in self.loggers:
            logger.log_batched_images(outputs, global_step, epoch, tag)

    def log_numbers(self, numbers: dict, global_step=None, epoch=None, tag=None):
        for logger in self.loggers:
            logger.log_numbers(numbers, global_step, epoch, tag)


class Printer(Logger):
    def __init__(self):
        self.on_main_thread = not dist.is_initialized() or torch.cuda.current_device() == 0

    def log_info(self, msg):
        formated_msg = f'{time.asctime(time.localtime())} - {msg}'
        if self.on_main_thread:
            print(formated_msg, flush=True)


class StandardLogger(Logger):
    def __init__(self, _ctx: Context):
        self.context = _ctx

    def setup(self, outputs_dir):
        self.outputs_dir = outputs_dir

        if not dist.is_initialized():
            suffix = ''
        else:
            suffix = f'_{dist.get_rank()}'
        self.name = self.context.exp_name + suffix

        self.on_main_thread = not dist.is_initialized() or torch.cuda.current_device() == 0

        # create file handler
        os.makedirs(self.outputs_dir, exist_ok=True)

        self.files = [
            open(os.path.join(self.outputs_dir,
                              f'log{suffix}.txt'), 'a')]

        if self.on_main_thread:
            os.makedirs(os.path.join(
                self.outputs_dir, 'css'), exist_ok=True)
            with open(os.path.join(self.outputs_dir, 'css', 'style.css'), 'w') as f:
                f.write(r'''
                    .styled-table {
                        border-collapse: collapse;
                        margin: 25px 0;
                        font-size: 0.9em;
                        font-family: sans-serif;
                        min-width: 400px;
                        box-shadow: 0 0 20px rgba(0, 0, 0, 0.15);
                    }
                    .styled-table thead tr {
                        background-color: #009879;
                        color: #ffffff;
                        text-align: left;
                    }
                    .styled-table th,
                    .styled-table td {
                        padding: 12px 15px;
                    }
                    .styled-table tbody tr {
                        border-bottom: 1px solid #dddddd;
                    }

                    .styled-table tbody tr:nth-of-type(even) {
                        background-color: #f3f3f3;
                    }

                    .styled-table tbody tr:last-of-type {
                        border-bottom: 2px solid #009879;
                    }

                    .styled-table tbody tr.active-row {
                        font-weight: bold;
                        color: #009879;
                    }
                ''')

    def __del__(self):
        for f in self.files:
            f.close()
        self.files.clear()

    def _print(self, msg: str):
        formated_msg = f'{time.asctime(time.localtime())} - {self.name} - {msg}'
        if self.on_main_thread:
            print(formated_msg, flush=True)
        for f in self.files:
            f.write(formated_msg + '\n')
            f.flush()

    def log_info(self, msg):
        self._print(msg)

    def log_batched_images(self, outputs, global_step=None, epoch=None, tag=None):
        if not dist.is_initialized() or torch.cuda.current_device() == 0:
            html = (f'<html><head><title>global_step={global_step}, epoch={epoch}</title>'
                    f'</head><body><table class="styled-table">')
            # write titles
            names = list(outputs.keys())
            html += '<thead><tr>'
            for name in names:
                html += f'<th>{name}</th>'
            html += '</tr></thead><tbody>'
            html += '<tr>'
            for name in names:
                image_batch = outputs[name]
                image_file_name = _tagged(f'{name}.png', tag)
                image_file_path = os.path.join(
                    self.outputs_dir, image_file_name)
                torchvision.utils.save_image(
                    image_batch, image_file_path, normalize=False, nrow=1)
                html += f'<td><img src="{image_file_name}" width="250"/></td>'
            html += '</tr>'
            html += '</tbody></table></body></html>'

            with open(os.path.join(self.outputs_dir, _tagged(f'images.html', tag)), 'w') as f:
                f.write(html)

    def log_numbers(self, numbers: dict, global_step=None, epoch=None, tag=None):
        line = ''
        for name, val in numbers.items():
            line += f'{_tagged(name, tag)}={val} '
        self._print(
            f'[global_step={global_step}, epoch={epoch}] {line}')


class TSVDataLogger(Logger):
    def __init__(self, _ctx=None):
        self.context = _ctx

    def setup(self, outputs_dir):
        self.outputs_dir = outputs_dir

        # create file handler
        os.makedirs(self.outputs_dir, exist_ok=True)

        if dist.is_initialized():
            self.rank = dist.get_rank()
            self.suffix = f'_{self.rank}'
        else:
            self.rank = None
            self.suffix = ''

        self.data_names = defaultdict(list)  # tag -> set[str]

    @staticmethod
    def _write_line(f, lst):
        for i, e in enumerate(lst):
            f.write(str(e))
            if i + 1 < len(lst):
                f.write('\t')
        f.write('\n')

    def log_numbers(self, numbers: dict, global_step=None, epoch=None, tag=None):
        if tag is None:
            fname = 'default'
        else:
            fname = tag.replace('\\', '_').replace('/', '_')
        with open(os.path.join(self.outputs_dir, fname + self.suffix + '.tsv'), 'a') as f:
            if set(self.data_names[tag]) != set(numbers.keys()):
                self.data_names[tag] = list(numbers.keys())
                self.__class__._write_line(
                    f, ['global_step', 'epoch'] + self.data_names[tag])

            values = [global_step, epoch] + [numbers[name]
                                             for name in self.data_names[tag]]
            self.__class__._write_line(f, values)


class TensorBoardLogger(Logger):
    def __init__(self, _ctx):
        self.context = _ctx

    def setup(self, outputs_dir):
        from torch.utils.tensorboard import SummaryWriter
        self.outputs_dir = outputs_dir

        if not dist.is_initialized() or dist.get_rank() == 0:
            if hasattr(self.context, '_tensorboard_log_dir'):
                log_dir = self.context._tensorboard_log_dir
            else:
                log_dir = self.outputs_dir
            os.makedirs(log_dir, exist_ok=True)
            self.writer = SummaryWriter(log_dir=log_dir)
        else:
            self.writer = None

    def log_numbers(self, numbers: dict, global_step=None, epoch=None, tag=None):
        if self.writer is not None:
            if global_step is not None:
                for name, val in numbers.items():
                    self.writer.add_scalar(
                        f'by_global_step/{_tagged(name, tag)}', val, global_step=global_step)
            if epoch is not None:
                for name, val in numbers.items():
                    self.writer.add_scalar(
                        f'by_epoch/{_tagged(name, tag)}', val, global_step=epoch)



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import numpy as np

import torch
import torch.nn as nn
import torch.distributed as dist

from typing import Optional, Dict, List
from packaging import version
import collections
import pickle as pk
import yaml
import shutil
import contextlib
import time

from .inference import BatchList
from .logger import Logger

_PYTORCH_VERSION_NEWER_THAN_180_ = version.parse(
    torch.__version__) >= version.parse('1.8.0')
if _PYTORCH_VERSION_NEWER_THAN_180_:
    def barrier():
        """The new version.
        """
        dist.barrier(device_ids=[torch.cuda.current_device()])
else:
    def barrier():
        """The old version.
        """
        dist.barrier()


FRAMEWORK_ROOT = os.path.abspath(os.path.dirname(__file__))


def _load_states(states_dir, logger: Logger,
                 global_step: Optional[int] = None,
                 epoch: Optional[int] = None):
    if not os.path.exists(states_dir):
        logger.log_info('%s does not exist' % states_dir)
        return None

    selected_filename = None
    if global_step is not None or epoch is not None:
        for n in os.listdir(states_dir):
            if not n.startswith('_') and n.endswith('.pth'):
                try:
                    step_str, epoch_str = n[:-len('.pth')].split('_')
                    _step = int(step_str)
                    _epoch = int(epoch_str)
                    if _step == global_step or _epoch == epoch:
                        selected_filename = os.path.join(states_dir, n)
                except:
                    continue
    if selected_filename is not None:
        logger.log_info('loading states from %s' % selected_filename)
        return torch.load(
            open(selected_filename, 'rb'),
            map_location={'cuda:0': f'cuda:{torch.cuda.current_device()}'})

    logger.log_info('locating the latest loadable state ...')
    state_files = [os.path.join(states_dir, n)
                   for n in os.listdir(states_dir)
                   if not n.startswith('_') and n.endswith('.pth')]
    state_files.sort(key=os.path.getmtime)

    for filename in state_files[::-1]:
        try:
            logger.log_info('loading states from %s' % filename)
            return torch.load(
                open(filename, 'rb'),
                map_location={'cuda:0': f'cuda:{torch.cuda.current_device()}'})
        except Exception as e:
            logger.log_info('Exception: %s' % e)
    logger.log_info('no valid state files found in %s' % states_dir)
    return None


def _save_states(state_dict: dict, states_dir: str, losses: Dict[str, float],
                 global_step: int, epoch: int, num_states_kept: Optional[int],
                 always_kept_steps: List[int], always_kept_epoches: List[int],
                 logger: Logger):

    assert num_states_kept is None or num_states_kept > 1
    if not os.path.exists(states_dir):
        os.makedirs(states_dir)

    # load existing records
    records_filename = os.path.join(states_dir, '_records.pth')
    if os.path.exists(records_filename):
        records = torch.load(records_filename)
    else:
        records = dict()

    # update records
    for loss_name, loss in losses.items():
        if loss_name not in records:
            records[loss_name] = (global_step, epoch, loss)
        else:
            _, _, prev_loss = records[loss_name]
            if loss < prev_loss:
                records[loss_name] = (global_step, epoch, loss)

    def _filename_overwritable(fn: str) -> bool:
        if not fn.endswith('.pth'):
            return False
        if fn.startswith('_'):
            return False

        _step = None
        _epoch = None
        try:
            step_str, epoch_str = fn[:-len('.pth')].split('_')
            _step = int(step_str)
            _epoch = int(epoch_str)
        except:
            return False

        if _step in always_kept_steps or _epoch in always_kept_epoches:
            return False
        for _, (s, e, _) in records.items():
            if (s, e) == (_step, _epoch):
                # this means (_step, _epoch) is stored in records
                return False
        return True

    overwritable_state_files = [os.path.join(states_dir, n)
                                for n in os.listdir(states_dir)
                                if _filename_overwritable(n)]
    overwritable_state_files.sort(key=os.path.getmtime)
    if num_states_kept is not None:
        logger.log_info(f'{len(overwritable_state_files)} state files are '
                        f'overwritable in folder {states_dir} '
                        f'will only keeping {num_states_kept} states')

    filename_to_save = os.path.join(
        states_dir, f'{global_step}_{epoch}.pth')
    if num_states_kept is None or len(overwritable_state_files) < num_states_kept:
        torch.save(state_dict, filename_to_save)
    else:
        earliest_file = overwritable_state_files[0]
        torch.save(state_dict, open(earliest_file, 'wb'))
        os.rename(earliest_file, filename_to_save)
    logger.log_info('state saved to %s' % filename_to_save)

    for loss_name, (s, e, v) in records.items():
        logger.log_info(('The best state for loss %s is in '
                        '%d_%d.pth with value %f') %
                        (loss_name, s, e, v))

    # save records
    torch.save(records, records_filename)
    logger.log_info('records saved to %s' % records_filename)


def to_descriptors(data):
    if isinstance(data, BatchList):
        return BatchList([to_descriptors(val) for val in data.values])
    if isinstance(data, collections.abc.Mapping):
        return {key: to_descriptors(val) for key, val in data.items()}
    elif isinstance(data, tuple) and hasattr(data, '_fields'):  # namedtuple
        return type(data)(*(to_descriptors(val) for val in data))
    elif isinstance(data, collections.abc.Sequence):
        return [to_descriptors(val) for val in data]
    elif isinstance(data, torch.Tensor):
        return f'torch.Tensor[dtype={data.dtype}, size={tuple(data.shape)}]'
    elif isinstance(data, np.ndarray):
        return f'np.ndarray[dtype={data.dtype}, size={tuple(data.shape)}]'
    else:
        return type(data)


def get_submodule(m: nn.Module, target: str) -> nn.Module:
    if target == "":
        return m

    atoms: List[str] = target.split(".")
    mod: nn.Module = m

    for item in atoms:
        if not hasattr(mod, item):
            raise AttributeError(mod._get_name() + " has no "
                                 "attribute `" + item + "`")
        mod = getattr(mod, item)
        if not isinstance(mod, nn.Module):
            raise AttributeError("`" + item + "` is not "
                                 "an nn.Module")
    return mod


def accumulate(model1: nn.Module, model2: nn.Module, decay: float):
    """Accumulate model2 to model1.

    model1 = model1 * decay + model2 * (1 - decay)
    """
    par1 = dict(model1.named_parameters())
    par2 = dict(model2.named_parameters())

    for k in par1.keys():
        par1[k].data.mul_(decay).add_(
            other=par2[k].data.to(par1[k].data.device),
            alpha=1 - decay)

    par1 = dict(model1.named_buffers())
    par2 = dict(model2.named_buffers())

    for k in par1.keys():
        if par1[k].data.is_floating_point():
            par1[k].data.mul_(decay).add_(
                other=par2[k].data.to(par1[k].data.device),
                alpha=1 - decay)
        else:
            par1[k].data = par2[k].data.to(par1[k].data.device)


def all_gather(data):
    """
    Run all_gather on arbitrary picklable data (not necessarily tensors)
    Args:
        data: any picklable object
    Returns:
        list[data]: list of data gathered from each rank
    """
    world_size = dist.get_world_size()
    if world_size == 1:
        return [data]

    # serialized to a Tensor
    buffer = pk.dumps(data)
    storage = torch.ByteStorage.from_buffer(buffer)
    tensor = torch.ByteTensor(storage).to("cuda")

    # obtain Tensor size of each rank
    local_size = torch.LongTensor([tensor.numel()]).to("cuda")
    size_list = [torch.LongTensor([0]).to("cuda") for _ in range(world_size)]
    dist.all_gather(size_list, local_size)
    size_list = [int(size.item()) for size in size_list]
    max_size = max(size_list)

    # receiving Tensor from all ranks
    # we pad the tensor because torch all_gather does not support
    # gathering tensors of different shapes
    tensor_list = []
    for _ in size_list:
        tensor_list.append(torch.ByteTensor(size=(max_size,)).to("cuda"))
    if local_size != max_size:
        padding = torch.ByteTensor(size=(max_size - local_size,)).to("cuda")
        tensor = torch.cat((tensor, padding), dim=0)
    dist.all_gather(tensor_list, tensor)

    data_list = []
    for size, tensor in zip(size_list, tensor_list):
        buffer = tensor.cpu().numpy().tobytes()[:size]
        data_list.append(pk.loads(buffer))

    return data_list


def all_gather_by_part(lst: list, size: Optional[int]) -> list:
    if dist.get_rank() == 0:
        print(f'local list length: {len(lst)}')

    if size is None:
        result = sum(all_gather(lst), [])
    else:
        def _get_subset(i, data, size) -> list:
            start = min(i * size, len(data))
            stop = min(start + size, len(data))
            return data[start:stop]
        # get max length
        length = torch.tensor(len(lst), dtype=torch.int32, device='cuda')
        dist.all_reduce(length, dist.ReduceOp.MAX)
        length = length.cpu().item()
        # print(f'max local length: {length}')
        result = []
        for i in range((length+size-1) // size):
            gathered = sum(all_gather(_get_subset(i, lst, size)), [])
            result += gathered

    if dist.get_rank() == 0:
        print(f'gathered list length: {len(result)}')
    return result


def path_is_parent(parent_path: str, child_path: str) -> str:
    # Smooth out relative path names, note: if you are concerned about symbolic links,
    # you should use os.path.realpath too
    parent_path = os.path.abspath(parent_path)
    child_path = os.path.abspath(child_path)

    # Compare the common path of the parent and child path with the common path of just
    # the parent path. Using the commonpath method on just the parent path will regularise
    # the path name in the same way as the comparison that deals with both paths,
    # removing any trailing path separator
    return os.path.commonpath([parent_path]) == os.path.commonpath([parent_path, child_path])


def get_connection_string(blob_root, container_name: str) -> str:
    blob_config = yaml.safe_load(
        open(os.path.join(blob_root, 'config', 'blob.yaml'), 'r'))
    return blob_config[container_name]['connection_string']


def get_query_string(blob_root, container_name: str) -> str:
    blob_config = yaml.safe_load(
        open(os.path.join(blob_root, 'config', 'blob.yaml'), 'r'))
    return blob_config[container_name]['query_string']


def download(blob_root: str, remote_path: str, local_path: str):
    """
    Args:
        remote_path (str): e.g ':blob/haya/xxx', or some regular path.
        local_path (str): a regular local path.
    """
    if remote_path.startswith(':blob/'):
        _, container_name, *path = remote_path.split('/')
        sig = get_query_string(blob_root, container_name)
        path = os.path.join(*path)

        print(f"Downloading {path} from container_name {container_name} ...")
        os.environ['AZCOPY_CRED_TYPE'] = 'Anonymous'
        azcopy_exe = os.path.join(blob_root, 'tool', 'aml', 'azcopy')
        os.system(f'{azcopy_exe} copy '
                  f'"https://facevcstandard.blob.core.windows.net/{container_name}/'
                  f'{path}?{sig}" '
                  f'"{local_path}" --overwrite=true --check-md5 FailIfDifferent '
                  f'--from-to=BlobLocal --recursive --trusted-microsoft-suffixes= --log-level=ERROR')
        os.environ['AZCOPY_CRED_TYPE'] = ''
        print('Done')

    else:
        shutil.copy(remote_path, local_path)


def deal_with_remote_file(remote_path: str, copy2local: bool, blob_root: str) -> str:
    if not copy2local and not remote_path.startswith(':blob/'):
        return remote_path

    if dist.is_initialized():
        local_model_path = os.path.basename(remote_path)
        if not os.path.exists(local_model_path) and torch.cuda.current_device() == 0:
            print(
                f'Copying {remote_path} to local directory {local_model_path}')
            download(blob_root, remote_path, local_model_path)
        barrier()
        return local_model_path
    else:
        local_model_path = os.path.basename(remote_path)
        if not os.path.exists(local_model_path):
            print(
                f'Copying {remote_path} to local directory {local_model_path}')
            download(blob_root, remote_path, local_model_path)
        return local_model_path


def dir_nonempty(dirname):
    # If directory exists and nonempty (ignore hidden files), prompt for action
    return os.path.isdir(dirname) and len([x for x in os.listdir(dirname) if x[0] != '.'])


def query_dir(d):
    if dir_nonempty(d):
        key = input(f'{d} is not empty, delete it? (y/N): ')
        if key == 'y':
            import shutil
            shutil.rmtree(d)
            return True
        else:
            return False
    else:
        return True


def try_mkdir(d):
    if not os.path.exists(d):
        os.mkdir(d)


def try_makedirs(d):
    if not os.path.exists(d):
        os.makedirs(d)


def read_name_list(filename):
    names = []
    for name in open(filename, 'r'):
        name = name.strip()
        if len(name) > 0:
            names.append(name)
    return names


def write_name_list(filename, names):
    with open(filename, 'w') as f:
        for name in names:
            f.write(name + '\n')


@contextlib.contextmanager
def record_time(name, collect_dict=None):
    start = time.time()
    yield
    stop = time.time()
    if collect_dict is None:
        print('time cost of %s is %f seconds' %
              (name, stop-start))
    else:
        collect_dict[f'{name}(s)'] = stop-start



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import random
import numpy as np
from copy import deepcopy
import torch
import torch.distributed as dist

from packaging import version

from ..builder import Builder


def _single_thread_run(gpu_rank, num_gpus, distributed_run: 'DistributedGPURun'):
    rank = distributed_run.node_rank * num_gpus + gpu_rank
    world_size = num_gpus * distributed_run.num_nodes

    if distributed_run.num_nodes == 1:
        init_method = 'tcp://localhost:10315'
    elif 'AZ_BATCHAI_PYTORCH_INIT_METHOD' in os.environ:
        init_method = os.environ['AZ_BATCHAI_PYTORCH_INIT_METHOD']
    elif 'MASTER_ADDR' in os.environ and 'MASTER_PORT' in os.environ:
        init_method = "tcp://%s:%s" % (
            os.environ['MASTER_ADDR'], os.environ['MASTER_PORT'])
    else:
        raise RuntimeError('Failed to create the init_method')

    dist.init_process_group(
        backend='nccl',
        init_method=init_method,
        rank=rank, world_size=world_size)

    print(f'DistributedGPURun: init_process_group: {rank}/{world_size}')

    torch.cuda.set_device(gpu_rank)
    seed = distributed_run.seed0 + rank
    torch.manual_seed(seed)
    random.seed(seed+100)
    np.random.seed(seed+200)

    # update the context for each thread
    builder: Builder = deepcopy(distributed_run.builder_snapshot)

    # the local_run is instantiated on each thread with renewed context
    local_run = builder.parse_cfg_object(distributed_run.local_run)

    local_run()

    # add a barrier to avoid `broken pipe` error
    # see https://github.com/pytorch/pytorch/issues/40633#issuecomment-651354635
    if version.parse(torch.__version__) >= version.parse('1.8.0'):
        dist.barrier(device_ids=[gpu_rank])
    else:
        dist.barrier()

    dist.destroy_process_group()


class DistributedGPURun:
    """Run another runnable object on distributed GPU nodes. 

    Supported by PyTorch utility.

    Args:
        local_run: The wrapped local run. 
            It runs on each local thread in each node.
        seed0: The intial seed for reproducibility.
    """

    @staticmethod
    def __parse_inputs__(local_run):
        return local_run

    def __init__(self, local_run, seed0: int = 123) -> None:
        if 'NODE_COUNT' in os.environ:
            self.num_nodes = int(os.environ['NODE_COUNT'])
        elif 'OMPI_COMM_WORLD_SIZE' in os.environ:
            self.num_nodes = int(os.environ['OMPI_COMM_WORLD_SIZE'])
        elif 'AZUREML_NODE_COUNT' in os.environ:
            self.num_nodes = int(os.environ['AZUREML_NODE_COUNT'])
        else:
            self.num_nodes = 1

        if 'RANK' in os.environ:
            self.node_rank = int(os.environ['RANK'])
        elif 'OMPI_COMM_WORLD_RANK' in os.environ:
            self.node_rank = int(os.environ['OMPI_COMM_WORLD_RANK'])
        elif 'AZ_BATCHAI_TASK_INDEX' in os.environ:
            self.node_rank = int(os.environ['AZ_BATCHAI_TASK_INDEX'])
        else:
            self.node_rank = 0

        self.local_run = local_run
        self.seed0 = seed0

        self.builder_snapshot = deepcopy(self.__class__.__builder__)

    def __call__(self):
        num_gpus = torch.cuda.device_count()
        # _single_thread_run(0, 1, self)
        torch.multiprocessing.spawn(
            _single_thread_run, args=(num_gpus, self), nprocs=num_gpus, join=True)



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Mapping, Type, Any
import os
import pickle
import numpy as np
import cv2
import zipfile
import io
from enum import Enum

from .dataset import Dataset


class Modality(Enum):
    CONFIG = 0
    IMAGE_PNG = 1
    IMAGE_JPG = 2
    IMAGE_LABEL = 3
    IMAGE_GRAY = 4


CONFIG = Modality.CONFIG
IMAGE_PNG = Modality.IMAGE_PNG
IMAGE_JPG = Modality.IMAGE_JPG
IMAGE_LABEL = Modality.IMAGE_LABEL
IMAGE_GRAY = Modality.IMAGE_GRAY


class Mode(Enum):
    READ = 0
    WRITE = 1


READ = Mode.READ
WRITE = Mode.WRITE


class _CompatibleUnpickler(pickle.Unpickler):
    def find_class(self, module_name: str, global_name: str) -> Any:
        if module_name.startswith('haya.'):
            module_name = 'blueprint.' + module_name[len('haya.'):]
        if module_name == 'blueprint.data' and global_name == 'Modality':
            module_name = 'blueprint.ml.storage'
        return super().find_class(module_name, global_name)


def _compatible_load(f):
    return _CompatibleUnpickler(f).load()


def _compatible_loads(bs):
    return _CompatibleUnpickler(io.BytesIO(bs)).load()


class Storage:
    """All storage classes should inherit this class.
    """

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def __del__(self):
        self.close()

    def allocate_key(self, key: str) -> None:
        pass

    def save(self, sample_name: str, key: str, value, modality: Modality) -> str:
        raise NotImplementedError()

    def load(self, sample_name: str, key: str, rep: str, modality: Modality):
        raise NotImplementedError()

    def save_config(self, config: dict):
        raise NotImplementedError()

    def load_config(self) -> dict:
        raise NotImplementedError()

    def close(self):
        pass


def _try_mkdir(s):
    if not os.path.exists(s):
        os.mkdir(s)


class FilesStorage(Storage):
    def __init__(self, root, mode: Mode):
        assert root is not None
        _try_mkdir(root)
        self.root = root

    def allocate_key(self, key):
        folder = os.path.join(self.root, key)
        _try_mkdir(folder)

    def save(self, sample_name, key, value, modality):
        if modality == Modality.IMAGE_PNG:
            assert value.ndim == 3
            cv2.imwrite(os.path.join(
                self.root, key, sample_name+'.png'), value[:, :, ::-1])
            return os.path.join(key, sample_name+'.png')
        elif modality == Modality.IMAGE_JPG:
            assert value.ndim == 3
            cv2.imwrite(os.path.join(
                self.root, key, sample_name+'.jpg'), value[:, :, ::-1])
            return os.path.join(key, sample_name+'.jpg')
        elif modality in {Modality.IMAGE_LABEL, Modality.IMAGE_GRAY}:
            assert value.ndim == 2
            cv2.imwrite(os.path.join(
                self.root, key, sample_name+'.png'), value)
            return os.path.join(key, sample_name+'.png')
        else:
            return value

    def load(self, sample_name, key, rep, modality):
        if modality in {Modality.IMAGE_JPG, Modality.IMAGE_PNG}:
            return cv2.cvtColor(cv2.imread(
                os.path.join(self.root, rep)), cv2.COLOR_BGR2RGB)
        elif modality in {Modality.IMAGE_LABEL, Modality.IMAGE_GRAY}:
            return cv2.imread(os.path.join(
                self.root, rep), cv2.IMREAD_GRAYSCALE)
        else:
            return rep

    def save_config(self, config):
        with open(os.path.join(self.root, 'config.pk'), 'wb') as f:
            pickle.dump(config, f)

    def load_config(self):
        with open(os.path.join(self.root, 'config.pk'), 'rb') as f:
            return _compatible_load(f)


class ZipStorage(Storage):
    def __init__(self, filename, mode: Mode, level: int = 0):
        if level > 0:
            compression = zipfile.ZIP_DEFLATED
        else:
            compression = zipfile.ZIP_STORED
        self.zipfile = zipfile.ZipFile(
            filename, 'r' if mode == Mode.READ else 'w',
            compression=compression, compresslevel=level)

    @staticmethod
    def _join(*cs):
        return '/'.join(cs)

    def save(self, sample_name, key, value, modality):
        if modality == Modality.IMAGE_PNG:
            assert value.ndim == 3
            ret, buf = cv2.imencode('.png', value[:, :, ::-1])
            assert ret
            path = ZipStorage._join(key, sample_name+'.png')
            self.zipfile.writestr(path, buf)
            return path
        elif modality == Modality.IMAGE_JPG:
            assert value.ndim == 3
            ret, buf = cv2.imencode('.jpg', value[:, :, ::-1])
            assert ret
            path = ZipStorage._join(key, sample_name+'.jpg')
            self.zipfile.writestr(path, buf)
            return path
        elif modality in {Modality.IMAGE_LABEL, Modality.IMAGE_GRAY}:
            assert value.ndim == 2
            ret, buf = cv2.imencode('.png', value)
            assert ret
            path = ZipStorage._join(key, sample_name+'.png')
            self.zipfile.writestr(path, buf)
            return path
        else:
            return value

    def load(self, sample_name, key, rep, modality):
        if modality in {Modality.IMAGE_JPG, Modality.IMAGE_PNG}:
            buf = self.zipfile.read(rep)
            return cv2.cvtColor(cv2.imdecode(
                np.frombuffer(buf, dtype=np.uint8), cv2.IMREAD_COLOR), cv2.COLOR_BGR2RGB)
        elif modality in {Modality.IMAGE_LABEL, Modality.IMAGE_GRAY}:
            buf = self.zipfile.read(rep)
            return cv2.imdecode(
                np.frombuffer(buf, dtype=np.uint8), cv2.IMREAD_GRAYSCALE)
        else:
            return rep

    def save_config(self, config):
        buf = pickle.dumps(config)
        self.zipfile.writestr('config.pk', buf)

    def load_config(self):
        buf = self.zipfile.read('config.pk')
        return _compatible_loads(buf)

    def close(self):
        self.zipfile.close()


class Frozen(Dataset):
    def __init__(self, filename, storage_type=None, storage_args=dict()):
        self.filename = filename
        self.storage_type = storage_type
        if self.storage_type is None:
            if os.path.isdir(self.filename):
                self.storage_type = FilesStorage
            elif self.filename.lower().endswith('.zip'):
                self.storage_type = ZipStorage
        self.storage = None

        self.storage_args = storage_args
        config = self.storage_type(
            self.filename, Mode.READ, **storage_args).load_config()
        self.data = config['data']
        self.modal = config['modal']
        self.extra_data_reps = config.get('extra_data', dict())
        self.extra_data_cache = None

    def setup(self):
        if self.storage is None:
            self.storage = self.storage_type(
                self.filename, Mode.READ, **self.storage_args)

    def extra_data(self):
        if self.extra_data_cache is None:
            self.setup()
            self.extra_data_cache = dict()
            for key, rep in self.extra_data_reps.items():
                modality = self.modal.get(key, Modality.CONFIG)
                self.extra_data_cache[key] = self.storage.load(
                    '_extra_data', key, rep, modality)
        return self.extra_data_cache

    def __len__(self):
        return len(self.data)

    def sample_name(self, index):
        return self.data[index]['sample_name']

    def __getitem__(self, index):
        self.setup()
        d = self.data[index]
        results = dict()
        for key, rep in d.items():
            modality = self.modal.get(key, Modality.CONFIG)
            results[key] = self.storage.load(
                d['sample_name'], key, rep, modality)
        return results

    def __del__(self):
        if self.storage is not None:
            self.storage.close()
        self.storage = None


def freeze(dataset: Dataset, filename: str, modal: Mapping[str, Modality],
           storage_type: Type[Storage] = ZipStorage,
           storage_args: Mapping[str, Any] = dict(),
           with_prog: bool = False, raise_errors: bool = False):
    r"""Freeze a dataset into a storage.

    Args:
        dataset: The dataset to freeze
        filename (str): The filename of the storage
        modal (dict): The modalities for each data tags
        storage_type: A subclass of `haya_data.Storage`, e.g `haya_data.ZipStorage`
        storage_args: The arguments for storage construction        
    """
    if with_prog:
        from tqdm.auto import tqdm
        p = tqdm(range(len(dataset)))
    else:
        p = range(len(dataset))

    if len(dataset) == 0:
        return

    with storage_type(filename, Mode.WRITE, **storage_args) as storage:
        assert isinstance(storage, Storage)
        for key in dataset[0].keys():
            storage.allocate_key(key)

        data = []
        for i in p:
            sample_name = dataset.sample_name(i)
            try:
                d = dataset[i]
            except InterruptedError:
                print('user interrupted')
                break
            except Exception as e:
                if raise_errors:
                    raise e
                print(f'error in sample: {sample_name}')
                print(f'error: {str(e)}')
                continue

            datum = {'sample_name': sample_name}
            for key, value in d.items():
                modality = modal.get(key, Modality.CONFIG)
                datum[key] = storage.save(sample_name, key, value, modality)
            data.append(datum)

        extra_data = dict()  # key -> rep
        for key, value in dataset.extra_data().items():
            modality = modal.get(key, Modality.CONFIG)
            extra_data[key] = storage.save('_extra_data', key, value, modality)

        print(
            f'in total {len(data)} out of {len(dataset)} samples will be stored')
        config = {'data': data, 'modal': modal,
                  'extra_data': extra_data}
        storage.save_config(config)



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Mapping, List, Dict, Optional, Any

import os
import torch
from torch import nn
import torch.distributed as dist
import pickle as pk

from .dataprocessor import DataProcessor
from .logger import Logger
from .task import Task
from .scorer import Scorer
from .trainer import Trainer, _interpret_ema_networks
from .util import _load_states


__all__ = ['Tester', 'tester_of_trainer']


def tester_of_trainer(trainer: Trainer, outputs_dir: str,
                      eval_dataprocessors: Optional[Mapping[str, DataProcessor]],
                      test_dataprocessors: Optional[Mapping[str, DataProcessor]],
                      logger: Optional[Logger] = None, epoch: Optional[int] = None):
    if eval_dataprocessors is None:
        eval_dataprocessors = trainer.eval_dataprocessors
    if test_dataprocessors is None:
        test_dataprocessors = dict()
    if logger is None:
        logger = trainer.logger
    return Tester(outputs_dir, trainer.states_dir,
                  eval_dataprocessors, test_dataprocessors,
                  trainer.networks, trainer.task_module, trainer.scorer,
                  logger, init_epoch=epoch)


class Tester:
    """A tester.
    """

    def __init__(self,
                 outputs_dir: str,
                 trainer_states_dir: str,
                 eval_dataprocessors: Mapping[str, DataProcessor],
                 test_dataprocessors: Mapping[str, DataProcessor],
                 networks: Mapping[str, nn.Module],
                 task_module: Task,
                 scorer: Scorer,
                 logger: Logger,
                 init_epoch: Optional[int] = None,
                 init_global_step: Optional[int] = None,
                 save_eval_results: bool = True):

        self.trainer_states_dir = trainer_states_dir
        self.outputs_dir = outputs_dir
        self.eval_dataprocessors = eval_dataprocessors
        self.test_dataprocessors = test_dataprocessors

        self.networks = _interpret_ema_networks(networks)
        self.task_module = task_module
        self.scorer = scorer

        self.logger = logger
        self.logger.setup(outputs_dir)
        self.logger.log_info(f'trainer_states_dir={trainer_states_dir}')

        self.init_global_step = init_global_step
        self.init_epoch = init_epoch

        self.global_step = None
        self.epoch = None

        self.save_eval_results = save_eval_results

    def __call__(self):
        if not dist.is_initialized():
            raise RuntimeError(
                "Requires torch.distributed package to be initialized")

        self.epoch = 0
        self.global_step = 0

        self._load_test_states()

        self.task_module.setup_networks(self.networks)
        task_module = nn.parallel.DistributedDataParallel(
            self.task_module.float().cuda(), device_ids=[torch.cuda.current_device()],
            find_unused_parameters=True)
        task_module.eval()

        for eval_name, eval_dataprocessor in self.eval_dataprocessors.items():
            self.logger.log_info(
                f'evaluating {eval_dataprocessor.num_steps_of_epoch(0)} '
                f'batches from {eval_name} ...')
            self.scorer.init_evaluation()
            for idx, result_data in enumerate(eval_dataprocessor.iterate_final_result(self.task_module)):
                self.scorer.evaluate(result_data)

                if self.save_eval_results:
                    save_dir = os.path.join(
                        self.outputs_dir, eval_name)
                    os.makedirs(save_dir, exist_ok=True)
                    with open(os.path.join(save_dir, f'{idx}_rank{dist.get_rank()}.pk'), 'wb') as f:
                        pk.dump(result_data, f)

            scores = self.scorer.finalize_evaluation()
            self.logger.log_numbers(scores,
                                    global_step=self.global_step,
                                    epoch=self.epoch, tag=f'eval.{eval_name}')

    def _load_test_states(self):
        states = _load_states(self.trainer_states_dir, self.logger,
                              global_step=self.init_global_step, epoch=self.init_epoch)
        if states is not None:
            self.epoch = states['epoch']
            self.global_step = states['global_step']
            self.logger.log_info(
                f'epoch={self.epoch}, global_step={self.global_step}')
            for net_name, net in self.networks.items():
                net.load_state_dict(states['networks'][net_name])
            del states
        else:
            raise RuntimeError(
                f'Failed to load any states from {self.trainer_states_dir}.')



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import operator
from typing import Mapping, MutableMapping, Any

import torch
import torch.distributed as dist


def _reduce_to_dict(d: MutableMapping[str, Any], name: str, val: Any,
                    reduce_op: 'dist.ReduceOp'):

    if reduce_op == dist.ReduceOp.SUM:
        reduce_op = operator.add
    elif reduce_op == dist.ReduceOp.PRODUCT:
        reduce_op = operator.mul
    elif reduce_op == dist.ReduceOp.MIN:
        reduce_op = min
    elif reduce_op == dist.ReduceOp.MAX:
        reduce_op = max

    if isinstance(val, torch.Tensor):
        if reduce_op == min:
            reduce_op = torch.min
        elif reduce_op == max:
            reduce_op = torch.max
    if name in d:
        d[name] = reduce_op(d[name], val)
    else:
        d[name] = val


def _reduce_all_distributed(d: Mapping[str, torch.Tensor],
                            reduce_op: 'dist.ReduceOp'):
    reduced = dict()
    for name, val in d.items():
        reduced_val = torch.tensor(val)
        dist.all_reduce(reduced_val, op=reduce_op)
        reduced[name] = reduced_val
    return reduced


def _simplify_tensor(t: torch.Tensor):
    if t.numel() == 1:
        return t.item()
    return t


class StatSummary:
    """Summary of all statistics.
    """

    def __init__(self, device=torch.device('cpu')):
        self.sums = dict()
        self.sum_sqs = dict()
        self.mins = dict()
        self.maxs = dict()
        self.count = 0
        self.device = device

    def update(self, batch_stat: Mapping[str, torch.Tensor]):
        """Update statistics from one batch.
        """

        batch_size = None
        for name, batch in batch_stat.items():
            batch = batch.detach().to(self.device)
            if batch.is_floating_point():
                batch = batch.float()

            _reduce_to_dict(self.sums, name, batch.sum([0]), dist.ReduceOp.SUM)
            _reduce_to_dict(self.sum_sqs, name, batch.pow(
                2).sum([0]), dist.ReduceOp.SUM)

            batch_min_v, _ = batch.min(0)
            _reduce_to_dict(self.mins, name, batch_min_v, dist.ReduceOp.MIN)

            batch_max_v, _ = batch.max(0)
            _reduce_to_dict(self.maxs, name, batch_max_v, dist.ReduceOp.MAX)

            if batch_size is not None:
                if batch_size != batch.size(0):
                    raise RuntimeError(
                        f'The first dimension of {name} ({batch.size(0)}) does not '
                        f'match the previously recorded batch size ({batch_size})')
            else:
                batch_size = batch.size(0)
        if batch_size is not None:
            self.count += batch_size

    def gather(self):
        """Gather statistics from all other threads.
        """

        self.sums = _reduce_all_distributed(self.sums, dist.ReduceOp.SUM)
        self.sum_sqs = _reduce_all_distributed(self.sum_sqs, dist.ReduceOp.SUM)
        self.mins = _reduce_all_distributed(self.mins, dist.ReduceOp.MIN)
        self.maxs = _reduce_all_distributed(self.maxs, dist.ReduceOp.MAX)
        count = torch.tensor(self.count)
        dist.all_reduce(count, op=dist.ReduceOp.SUM)
        self.count = count.item()

    def get_averages(self) -> Mapping[str, Any]:
        return {name: _simplify_tensor(s/self.count) for name, s in self.sums.items()}

    def get_stddevs(self) -> Mapping[str, Any]:
        avgs = self.get_averages()
        avg_sqs = {name: _simplify_tensor(s/self.count)
                   for name, s in self.sum_sqs.items()}
        stddevs = dict()
        for name, _ in self.sums.items():
            stddevs[name] = (avg_sqs[name] - avgs[name] ** 2) ** 0.5
        return stddevs



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import pathlib
import yaml
import functools


@functools.lru_cache()
def _load_user_config():
    home_dir = str(pathlib.Path.home())
    config_path = os.path.join(home_dir, 'blueprint_cfg.yaml')
    if not os.path.exists(config_path):
        return dict()
    return yaml.safe_load(open(config_path, 'r'))


def get_user_config(key: str, default_val=None):
    return _load_user_config().get(key, default_val)



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Optional
from dataclasses import dataclass


@dataclass
class Context:
    exp_name: Optional[str] = None  # experiment name

    blob_root: Optional[str] = None  # blob root path
    copy2local: Optional[bool] = False  # copy remote data to local machine?

    num_nodes: Optional[int] = 1
    node_rank: Optional[int] = 0

    debug_loss: bool = False
    detect_anomaly: bool = True



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import shutil
from typing import Optional, Tuple, List, Union

import numpy as np
import random
import math
import torch
from torch.utils.data import (
    DataLoader, Sampler, DistributedSampler, RandomSampler, SequentialSampler)
from torch import distributed as dist

from .dataset import (
    Dataset, DatasetRepeatingConstantValue, ConcatDataset, Subset)
from .augmenters import augment_each
from .inference import collate, decollate
from .storage import Frozen
from .augmenters import AugLike


from .util import barrier
from .task import Task, ForwardFlags

from ..context import Context


__all__ = ['DataProcessor', 'DataSource']


class _WorkerInitFunction:
    def __init__(self, rank) -> None:
        self.rank = rank

    def __call__(self, worker_id):
        seed = worker_id * 1000 + self.rank
        torch.manual_seed(seed)
        random.seed(seed+100)
        np.random.seed(seed+200)


def _get_data_loader_and_sampler(
        dataset: Dataset, batch_size: int,
        randomize: bool = False, num_workers: int = 0,
        validity_tag: str = 'valid') -> Tuple[DataLoader, Sampler]:
    """Get the DataLoader and Sampler from a dataset.
    """
    if dist.is_initialized():
        world_size = dist.get_world_size()
    else:
        world_size = 1

    dataset = dataset.make_evenly_divisible(
        world_size * batch_size, validity_tag=validity_tag)
    assert len(dataset) % (world_size * batch_size) == 0

    if dist.is_initialized():
        sampler = DistributedSampler(
            dataset, shuffle=randomize, drop_last=False)
        rank = dist.get_rank()
    else:
        if randomize:
            sampler = RandomSampler(dataset)
        else:
            sampler = SequentialSampler(dataset)
        rank = 0

    data_loader = DataLoader(
        dataset, batch_size=batch_size, sampler=sampler,
        collate_fn=collate, drop_last=False,
        num_workers=num_workers,
        pin_memory=True, worker_init_fn=_WorkerInitFunction(rank))
    return data_loader, sampler


def _load_frozen_dataset(filepath, copy2local=True, local_filepath=None,
                         subset_ids=None):
    if copy2local:
        if local_filepath is None:
            local_filepath = os.path.basename(filepath)

        # dist
        if not dist.is_initialized():
            if not os.path.exists(local_filepath):
                print(
                    f'Copying {filepath} to local directory {local_filepath}')
                shutil.copy(filepath, local_filepath)
        else:
            if torch.cuda.current_device() == 0:
                if not os.path.exists(local_filepath):
                    print(
                        f'Copying {filepath} to local directory {local_filepath}')
                    shutil.copy(filepath, local_filepath)
            barrier()
        filepath = local_filepath
    dataset = Frozen(filepath)
    if subset_ids is not None:
        dataset = Subset(dataset, subset_ids)
    return dataset


class DataSource:
    """Data source representation.

    Args:
        data_path (Union[str, List[str]]): A frozen zip path, or a list of paths.
        random_ratio (Optional[float]): When random_ratio < 1.0, a subset of 
            data with random_ratio would be randomly sampled from the original dataset. 
            However, the data size is unchanged through random repeating.
        subset_ids (Optional[List[int]]): Subset data indices.
        augmentations (AugLike): Data transforms before feeding into network.
        post_augmentations (AugLike): Data transforms appended to the network outputs.

    """

    def __init__(self, data_path: Union[str, List[str]],
                 random_ratio: Optional[float] = None,
                 random_subsample_seed: int = 123,
                 augmentations: AugLike = None,  post_augmentations: AugLike = None,
                 subset_ids: Optional[List[int]] = None,
                 _ctx: Optional[Context] = None):
        if isinstance(data_path, str):
            data_path = [data_path]
        self.data_path = data_path

        dataset = ConcatDataset([
            _load_frozen_dataset(os.path.join(
                _ctx.blob_root, p), copy2local=_ctx.copy2local)
            for p in self.data_path])

        if random_ratio is not None:
            rnd = random.Random(random_subsample_seed)
            orig_len = len(dataset)
            selected_ids = list(range(orig_len))
            rnd.shuffle(selected_ids)
            del selected_ids[int(orig_len*random_ratio):]
            assert len(set(selected_ids)) == int(orig_len*random_ratio)
            # repeat the selected ids to orig_len
            selected_ids = selected_ids * math.ceil(orig_len/len(selected_ids))
            del selected_ids[orig_len:]
            assert len(selected_ids) == orig_len
            assert len(set(selected_ids)) == int(orig_len*random_ratio)
            dataset = Subset(dataset, selected_ids)

        if subset_ids is not None:
            dataset = Subset(dataset, subset_ids)

        self.dataset = dataset
        self.augmentations = augmentations
        self.dataset_augmented = self.dataset.augment(self.augmentations)
        self.post_augmentations = post_augmentations


class DataProcessor:
    """Data processors.

    DataProcessor processes data following the pipleline as below:
    ```
        Data loaded from data_path 
            -> augmentations 
            -> collate as input data batches in DataLoader [A]
            -> task_module generates output data batches
            -> decollate data batches as individual data samples
            -> post augmentations
            -> final results [B]
    ```

    Use `iterate_input` to iterate the data batches in [A], 
    use `iterate_final_result` to iterate the final results in [B].

    """

    def __init__(self, data_src: DataSource, batch_size: int,
                 randomize: bool, num_workers: int = 2,
                 validity_tag: str = 'valid',
                 repeats_constant: bool = False) -> None:
        self.data_src = data_src
        self.validity_tag = validity_tag

        dataset_augmented = self.data_src.dataset_augmented
        if repeats_constant:
            dataset_augmented = DatasetRepeatingConstantValue(
                dataset_augmented[0], len(dataset_augmented))

        self.data_loader, self.sampler = _get_data_loader_and_sampler(
            dataset_augmented, batch_size=batch_size,
            randomize=randomize, num_workers=num_workers, validity_tag=validity_tag)

    def num_steps_of_epoch(self, epoch_idx: int = 0) -> int:
        if self.data_loader is None:
            self.load()
        if hasattr(self.sampler, 'set_epoch'):
            self.sampler.set_epoch(epoch_idx)
        return len(self.data_loader)

    def iterate_input(self, epoch_idx: int = 0):
        """Iterate the batched data input to nn.Modules. 

        By default most data will be torch.Tensor.
        """
        if self.data_loader is None:
            self.load()
        if hasattr(self.sampler, 'set_epoch'):
            self.sampler.set_epoch(epoch_idx)
        return iter(self.data_loader)

    def iterate_final_result(self, task_module: Task, epoch_idx: int = 0):
        """Iterate the final non-batched data after post_augmentation.

        By default most data will be np.ndarray.
        """
        with torch.no_grad():
            for batch_data in self.iterate_input(epoch_idx):
                _, _, batch_result, _ = task_module(
                    batch_data, ForwardFlags(with_losses=False, with_outputs=True,
                                             with_images=False))
                full_batch = {**batch_data, **batch_result}
                for single_result in augment_each(decollate(full_batch).values,
                                                  self.data_src.post_augmentations):
                    is_valid_data = single_result.get(self.validity_tag, True)
                    if is_valid_data:
                        yield single_result



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import collections
from typing import List, Optional, Tuple, Dict, Set, Mapping, Union, Any, Callable

import numpy as np
from copy import deepcopy

from ..util import record_time
from ..inference import NonStackable
from ..dataset import Dataset


class AugBase(object):
    r"""Base class for all augmenters
    """

    def process(self, data: Any):
        raise NotImplementedError()


AugLike = Union[AugBase, list, Callable[..., Any], None]


class AugmentedDataset(Dataset):
    """The augmented dataset.

    Args:
        dataset: The original dataset before augmentation
        aug: The augmenter
    """

    def __init__(self, dataset: Dataset, aug: AugBase):
        self.dataset = dataset
        self.aug = aug

    def __len__(self):
        return len(self.dataset)

    def sample_name(self, index):
        return self.dataset.sample_name(index)

    def __getitem__(self, index):
        return self.aug.process(self.dataset[index])

    def extra_data(self) -> dict:
        return self.dataset.extra_data()


def _wrap_aug(aug: AugLike) -> AugBase:
    if isinstance(aug, list):
        aug = Sequential(aug)
    elif callable(aug):
        aug = Lambda(aug)
    elif aug is None:
        aug = DoNothing()
    assert isinstance(aug, AugBase)
    return aug


def augment(dataset: Dataset, aug: AugLike):
    r"""Augment a dataset with given augmenter

    Args:
        dataset: The original dataset
        aug: The augmenter
    """
    return AugmentedDataset(dataset, _wrap_aug(aug))


def augment_each(data_seq: collections.abc.Sequence, aug: AugLike):
    r"""Augment a given data sequence with given augmenter

    Args:
        data_seq: The input data sequence
        aug: The augmenter
    """
    aug = _wrap_aug(aug)
    for data in data_seq:
        yield aug.process(data)


class DoNothing(AugBase):
    r"""An augmenter that does nothing
    """

    def process(self, data):
        return data


class Sequential(AugBase):
    r"""A sequence of augmenters that applies in a one-by-one style
    """

    def __init__(self, sub_augs: List[AugLike]):
        self.sub_augs = [_wrap_aug(aug) for aug in sub_augs]

    def process(self, data):
        for aug in self.sub_augs:
            data = aug.process(data)
        return data


class Benchmark(AugBase):
    """An augmenter to measure time cost

    Args:
        name (str): A name for printing
        sub_aug: The augmenter whose time to measure
    """

    def __init__(self, name: str, sub_aug: AugLike):
        self.name = name
        self.sub_aug = _wrap_aug(sub_aug)

    def process(self, data):
        with record_time(self.name):
            ret = self.sub_aug.process(data)
        return ret


class Lambda(AugBase):
    r"""An augmenter that warps a functor within
    """

    def __init__(self, func):
        self.func = func

    def process(self, data):
        if isinstance(data, tuple):
            return self.func(*data)
        else:
            return self.func(data)


class With(AugBase):
    """An augmenter that wraps another augmenter with both input and output tag names
    """

    def __init__(self,
                 input_tags: Union[str, List[str], None] = None,
                 output_tags: Union[str, List[str], None] = None,
                 aug: AugLike = None,
                 tags_str: Optional[str] = None):

        if tags_str is not None:
            if not (input_tags is None and output_tags is None):
                raise RuntimeError(
                    'When tag_str is given, input_tags '
                    'and output_tags should both be None.')
            input_tags_str, output_tags_str = tags_str.split('->')

            def _interpret_tags_str(tags_str):
                tags = [s.strip() for s in tags_str.split(',')]
                if len(tags) == 1:
                    tags = tags[0]
                return tags
            input_tags = _interpret_tags_str(input_tags_str)
            output_tags = _interpret_tags_str(output_tags_str)

        self.input_tags = input_tags
        self.output_tags = output_tags

        self.aug = _wrap_aug(aug)

    @staticmethod
    def _get_value(data, tag):
        if isinstance(tag, str):
            return data[tag]
        return tag

    def process(self, data):
        if not isinstance(data, dict):
            raise RuntimeError(
                f'Please input dict data into the `With` augmenter.')

        if isinstance(self.input_tags, tuple):
            result = self.aug.process(
                tuple([__class__._get_value(data, tag)
                       for tag in self.input_tags]))
        elif isinstance(self.input_tags, list):
            result = self.aug.process([
                __class__._get_value(data, tag)
                for tag in self.input_tags])
        else:
            result = self.aug.process(
                __class__._get_value(data, self.input_tags))

        if isinstance(self.output_tags, str):
            return {**data, self.output_tags: result}
        elif isinstance(self.output_tags, (tuple, list)):
            assert isinstance(result, (tuple, list))
            assert len(self.output_tags) == len(result)
            result = {tag: re for tag, re in zip(self.output_tags, result)}
            return {**data, **result}
        else:
            raise RuntimeError(
                f'Unsupported output_tags type: {type(self.output_tags)}.')


class Rename(AugBase):
    """An augmenter that renames tags
    """

    def __init__(self, tag_names_mapping: Mapping[str, str]):
        self.tag_names_mapping = tag_names_mapping

    def process(self, data):
        out_data = dict()
        for tag_name, d in data.items():
            tag_name = self.tag_names_mapping.get(tag_name, tag_name)
            out_data[tag_name] = d
        return out_data


class DeepCopy(AugBase):
    """An augmenter that deepcopies tags
    """

    def __init__(self, tag_names_mapping: Mapping[str, str]):
        self.tag_names_mapping = tag_names_mapping

    def process(self, data):
        out_data = deepcopy(data)
        for tag_name, d in data.items():
            if tag_name in self.tag_names_mapping:
                out_data[self.tag_names_mapping[tag_name]] = d
        return out_data


class If(AugBase):
    """ condition:bool, input_value -> output_value
    """

    def __init__(self, then_branch: AugLike, else_branch: AugLike = None):
        self.then_branch = _wrap_aug(then_branch)
        self.else_branch = _wrap_aug(else_branch)

    def process(self, data):
        assert isinstance(data, tuple)
        cond, input_value = data
        assert isinstance(cond, bool)
        if cond:
            return self.then_branch.process(input_value)
        else:
            return self.else_branch.process(input_value)


class IfExist(AugBase):
    """ execute then_aug only if tags are all available.
    """

    def __init__(self, tags: Union[str, Set[str]], then_branch: AugLike, else_branch: AugLike = None):
        if isinstance(tags, str):
            tags = {tags}
        self.tags = set(tags)
        self.then_branch = _wrap_aug(then_branch)
        self.else_branch = _wrap_aug(else_branch)

    def process(self, data: Dict[str, Any]):
        if self.tags <= data.keys():
            return self.then_branch.process(data)
        else:
            return self.else_branch.process(data)


class AttachConstData(AugBase):
    r"""An augmenter that attaches a constant as new tag

    Example:

        >>> # you can attach any data using this code:
        >>> ds = ds.augment(AttachConstData('new_tag_name', const_data))

    """

    def __init__(self, tag_name: str, const_data: Any):
        self.tag_name = tag_name
        self.const_data = const_data

    def process(self, data):
        if self.tag_name in data:
            raise RuntimeWarning(
                f'existing tags already include the "{self.tag_name}", '
                f'whose data will be overwritten by AttachConstData')
        return {self.tag_name: self.const_data, **data}


class AttachListData(AugBase):
    r"""An augmenter that attaches a list as new tag

    Example:

        >>> # if list_data = [new_data1, new_data2, ...]
        >>> # then you can attach these data using this code:
        >>> ds = ds.augment(AttachListData('new_tag_name', list_data))

    """

    def __init__(self, tag_name: str, list_data: List[Any]):
        self.tag_name = tag_name
        self.list_data = list_data

    def process(self, data):
        assert isinstance(data, dict)
        if 'index' not in data:
            raise RuntimeError(
                'existing tags do not include "index", '
                'call .attach_index_as_tag() before using AttachListData')
        if self.tag_name in data:
            raise RuntimeWarning(
                f'existing tags already include the "{self.tag_name}", '
                f'whose data will be overwritten by AttachListData')
        return {self.tag_name: self.list_data[data['index']], **data}


class AttachDictData(AugBase):
    r"""An augmenter that attaches a dict as new tag

    Example:

        >>> # if dict_data = {sample_name1:new_data1, sample_name2:new_data2, ...}
        >>> # then you can attach these data using this code:
        >>> ds = ds.augment(AttachDictData('new_tag_name', dict_data))

    """

    def __init__(self, tag_name: str, dict_data: Dict[str, Any]):
        self.tag_name = tag_name
        self.dict_data = dict_data

    def process(self, data):
        assert isinstance(data, dict)
        if 'sample_name' not in data:
            raise RuntimeError(
                'existing tags do not include "sample_name", '
                'call .attach_sample_name_as_tag() before using AttachDictData')
        if self.tag_name in data:
            raise RuntimeWarning(
                f'existing tags already include the "{self.tag_name}", '
                f'whose data will be overwritten by AttachDictData')
        return {self.tag_name: self.dict_data[data['sample_name']], **data}


class Filter(AugBase):
    """ dict -> dict
    """

    def __init__(self, tags, strict=False):
        self.tags = tags
        self.strict = strict

    def process(self, data):
        assert isinstance(data, dict)
        if self.strict:
            return {name: data[name] for name in self.tags}
        else:
            return {name: data[name] for name in self.tags if name in data}


class Maybe(AugBase):
    """ randomly perform the augmenter given a probability """

    def __init__(self, prob: float, then_branch: AugLike, else_branch: AugLike = None):
        self.prob = prob
        self.then_branch = _wrap_aug(then_branch)
        self.else_branch = _wrap_aug(else_branch)

    def process(self, data):
        selected = (np.random.uniform() <= self.prob)
        if selected:
            data = self.then_branch.process(data)
        else:
            data = self.else_branch.process(data)
        return data


class RandomSwitch(AugBase):
    """ randomly pick one of the multiple branches with given probabilities """

    def __init__(self, prob_branches: List[Tuple[float, AugLike]]):
        self.probs = [branch[0] for branch in prob_branches]
        self.branches = [_wrap_aug(branch[1]) for branch in prob_branches]

    def process(self, data):
        selected = np.random.choice(len(self.probs), p=self.probs)
        return self.branches[selected].process(data)


class RandomPick(AugBase):
    """ randomly pick an element from a container 

        a list -> an element
    """

    def process(self, data):
        return np.random.choice(data)


class MakeNonStackable(AugBase):
    """ make this data non-stackable in collate """

    def process(self, data):
        if isinstance(data, NonStackable):
            return data
        else:
            return NonStackable(data)


class UnwrapNonStackable(AugBase):
    """ get the original data from the non-stackable wrapper """

    def process(self, data):
        if isinstance(data, NonStackable):
            return data.value
        else:
            return data



# MIT License

# Copyright (c) 2021 Hao Yang (yanghao.alexis@foxmail.com)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
The machine learning tools.
"""
from .util import barrier, all_gather, all_gather_by_part
from .distributed import DistributedGPURun
from .logger import Logger, Loggers, Printer, StandardLogger, TSVDataLogger, TensorBoardLogger
from .scorer import Scorer, MultipleScorers
from .task import Task, ForwardFlags
from .dataprocessor import DataProcessor, DataSource
from .trainer import EMAConfig, Trainer
from .tester import Tester, tester_of_trainer
from .stats import StatSummary


from .dataset import (Dataset, Subset, ConcatDataset, DatasetRepeatingConstantValue,
                      Split, ALL, TRAIN, VAL, TEST, REFINED_TEST, TOY)
from .storage import Storage, FilesStorage, ZipStorage, Frozen, freeze, Modality, Mode
from .inference import BatchList, NonStackable, collate, decollate
from .sampler import DistributedRandomMixSampler, UnifiedDistributedSampler

from . import augmenters as aug
from .augmenters import augment, augment_each
from . import viz
